# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Callable, Optional

import jax
import jax.numpy as jnp

from ott.math import _lbfgs as lbfgs

__all__ = ["legendre"]


def legendre(
    fun: Callable[[jnp.ndarray], jnp.ndarray],
    **kwargs: Any,
) -> Callable[[jnp.ndarray, Optional[jnp.ndarray], Any], jnp.ndarray]:
  """Legendre (Fenchel) transform of a function.

  The solution is computed numerically using L-BFGS.

  Args:
    fun: A function to be transformed, must be convex for the transform
      to be properly defined.
    kwargs: Keyword arguments for :func:`~ott.math.lbfgs`, e.g. maximal
      iterations ``max_iters``, convergence tolerance ``tol`` or
      :func:`optax.lbfgs` arguments.

  Returns:
    A function that computes numerically the Legendre transform of
    the ``fun`` at a given point.
  """

  def fun_star(
      x: jnp.ndarray,
      x_init: Optional[jnp.ndarray] = None,
  ) -> float:
    """Runs optimization to compute the Legendre transform of ``fun`` at ``x``.

    Args:
      x: Array of shape ``[d,]`` where to evaluate the function.
      x_init: Initialization for optimization, of the same size of ``x``.
        If :obj:`None`, use ``x``.

    Returns:
        The Legendre transform of the ``fun`` evaluated at ``x``.
    """
    x_init = x if x_init is None else x_init

    def mod_fun(z: jnp.ndarray) -> float:
      """Conjugate maximizes <x,z> - fun(z), here minimize fun(z) - <x,z>."""
      return fun(z) - jnp.dot(x, z)

    z, _ = lbfgs.lbfgs(fun=mod_fun, x_init=x_init, **kwargs)
    # Flip sign again to revert to maximization convention, stop gradient.
    return -mod_fun(jax.lax.stop_gradient(z))

  return fun_star



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
from typing import Optional, Tuple, Union

import jax
import jax.numpy as jnp
import jax.random as jr

from ott.solvers.linear import semidiscrete

__all__ = ["SemidiscreteDataloader"]


@dataclasses.dataclass(frozen=False, repr=False)
class SemidiscreteDataloader:
  """Semidiscrete dataloader.

  This dataloader samples from the continuous source distribution and
  couples it with the discrete target distribution. It returns aligned tuples of
  ``(source, target)`` arrays of shape ``[batch, ...]``.

  Args:
    rng: Random number seed used for sampling from the source distribution.
    sd_out: Semidiscrete output object storing a precomputed OT solution between
      the (continuous) source distribution and the dataset of interest.
    batch_size: Batch size.
    epsilon: Epsilon regularization. If :obj:`None`, use the one stored
      in the :attr:`geometry <ott.solvers.linear.semidiscrete.SemidiscreteOutput.geom>`.
    epsilon: Epsilon regularization. In the context of this class, this epsilon
      value can be interpreted exclusively as a softmax temperature.
      If :obj:`None`, use the one stored in the
      :attr:`geometry <ott.solvers.linear.semidiscrete.SemidiscreteOutput.geom>`
      which was used to compute the potential stored in the ``sd_out``.
    subset_size_threshold: Threshold above which to sample from a subset of the
      coupling matrix. Only applicable when the problem is :meth:`entropically
      regularized <ott.geometry.semidiscrete_pointcloud.SemidiscretePointCloud.is_entropy_regularized>`.
      If :obj:`None`, don't subset the coupling matrix.
    subset_size: Size of the subset of the coupling matrix. This will
      subset a coupling of shape ``[batch, m]`` to ``[batch, subset_size]``
      using the :func:`~jax.lax.top_k` values if ``m > subset_size_threshold``.
    return_indices: Whether to return, in addition to paired source and target
      data points, the indices corresponding to the selected target data points.
    out_shardings: Output shardings for the aligned batch.
  """  # noqa: E501
  rng: jax.Array
  sd_out: semidiscrete.SemidiscreteOutput
  batch_size: int
  epsilon: Optional[float] = None
  subset_size_threshold: Optional[int] = None
  subset_size: Optional[int] = None
  return_indices: bool = False
  out_shardings: Optional[jax.sharding.Sharding] = None

  def __post_init__(self) -> None:
    _, m = self.sd_out.geom.shape
    assert self.batch_size > 0, \
      f"Batch size must be positive, got {self.batch_size}."

    if self.subset_size_threshold is not None:
      assert 0 < self.subset_size_threshold < m, \
        f"Subset threshold must be in (0, {m}), " \
        f"got {self.subset_size_threshold}."
      assert 0 < self.subset_size < m, \
        f"Subset size must be in (0, {m}), got {self.subset_size}."

    self._rng_it: Optional[jax.Array] = None
    self._sample_fn = jax.jit(
        _sample,
        out_shardings=self.out_shardings,
        static_argnames=[
            "batch_size",
            "epsilon",
            "subset_size_threshold",
            "subset_size",
            "return_indices",
        ],
    )

  def __iter__(self) -> "SemidiscreteDataloader":
    """Return self."""
    self._rng_it = self.rng
    return self

  def __next__(
      self
  ) -> Union[Tuple[jax.Array, jax.Array], Tuple[jax.Array, jax.Array,
                                                jax.Array]]:
    """Sample from the source distribution and match it with the data.

    Returns:
      A tuple of samples and data, arrays of shape ``[batch, ...]`` and
      optionally the sampled target indices of shape ``[batch,]``.
    """
    assert self._rng_it is not None, "Please call `iter()` first."
    self._rng_it, rng_sample = jr.split(self._rng_it, 2)
    return self._sample_fn(
        rng_sample,
        self.sd_out,
        self.batch_size,
        self.epsilon,
        self.subset_size_threshold,
        self.subset_size,
        self.return_indices,
    )


def _sample(
    rng: jax.Array,
    out: semidiscrete.SemidiscreteOutput,
    batch_size: int,
    epsilon: Optional[float],
    subset_size_threshold: Optional[int],
    subset_size: int,
    return_indices: bool,
) -> Union[Tuple[jax.Array, jax.Array], Tuple[jax.Array, jax.Array, jax.Array]]:
  rng_sample, rng_tmat = jr.split(rng, 2)
  out_sampled = out.sample(rng_sample, batch_size, epsilon=epsilon)

  if isinstance(out_sampled, semidiscrete.HardAssignmentOutput):
    tgt_idx = out_sampled.paired_indices[1]
  else:
    tgt_idx = _sample_from_coupling(
        rng_tmat,
        out_sampled.matrix,
        subset_size_threshold=subset_size_threshold,
        subset_size=subset_size,
        axis=1,
    )

  src = out_sampled.geom.x
  tgt = out_sampled.geom.y[tgt_idx]
  return (src, tgt, tgt_idx) if return_indices else (src, tgt)


def _sample_from_coupling(
    rng: jax.Array,
    coupling: jax.Array,
    *,
    subset_size_threshold: Optional[int],
    subset_size: int,
    axis: int,
) -> jax.Array:
  assert axis in (0, 1), axis
  n, m = coupling.shape
  sampling_size = m if axis == 1 else n

  if subset_size_threshold is None or sampling_size <= subset_size_threshold:
    return jr.categorical(rng, jnp.log(coupling), axis=axis)

  oaxis = 1 - axis
  top_k_fn = jax.vmap(jax.lax.top_k, in_axes=[oaxis, None], out_axes=oaxis)
  coupling, idx = top_k_fn(coupling, subset_size)

  expected_shape = (subset_size, m) if axis == 0 else (n, subset_size)
  assert coupling.shape == expected_shape, (coupling.shape, expected_shape)

  sampled_idx = jr.categorical(rng, jnp.log(coupling), axis=axis)
  if axis == 0:
    return idx[sampled_idx, jnp.arange(m)]
  return idx[jnp.arange(n), sampled_idx]



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import jax
import jax.numpy as jnp
import jax.random as jr
import numpy as np
import scipy.sparse as sp

import matplotlib.colors as mcolors
import matplotlib.patches as ptc
import matplotlib.pyplot as plt
from matplotlib import animation

from ott.experimental import mmsinkhorn
from ott.geometry import pointcloud
from ott.solvers.linear import sinkhorn, sinkhorn_lr
from ott.solvers.quadratic import gromov_wasserstein

# TODO(michalk8): make sure all outputs conform to a unified transport interface
Transport = Union[sinkhorn.SinkhornOutput, sinkhorn_lr.LRSinkhornOutput,
                  gromov_wasserstein.GWOutput]

__all__ = ["Plot", "PlotMM", "transport_animation"]


class Plot:
  """Plot an optimal transport map between two point clouds.

  This object can either plot or update a plot, to create animations as a
  :class:`~matplotlib.animation.FuncAnimation`, which can in turned be saved to
  disk at will. There are two design principles here:

  #. we do not rely on saving to/loading from disk to create animations
  #. we try as much as possible to disentangle the transport problem from
     its visualization.

  We use 2D scatter plots by default, relying on PCA visualization for d>3 data.
  This step requires a conversion to a numpy array, in order to compute leading
  singular values. This tool is therefore not designed having performance in
  mind.

  Args:
    fig: Specify figure object. Created by default
    ax: Specify axes objects. Created by default
    threshold: value below which links in transportation matrix won't be
      plotted. This value should be negative when using animations.
    scale: scale used for marker plots.
    show_lines: whether to show OT lines, as described in ``ot.matrix`` argument
    cmap: color map used to plot line colors.
    scale_alpha_by_coupling: use or not the coupling's value as proxy for alpha
    alpha: default alpha value for lines.
    title: title of the plot.
  """

  def __init__(
      self,
      fig: Optional["plt.Figure"] = None,
      ax: Optional["plt.Axes"] = None,
      threshold: float = -1.0,
      scale: int = 200,
      show_lines: bool = True,
      cmap: str = "cool",
      scale_alpha_by_coupling: bool = False,
      alpha: float = 0.7,
      title: Optional[str] = None,
      xlim: Optional[List[float]] = None,
      ylim: Optional[List[float]] = None,
  ):
    if ax is None and fig is None:
      fig, ax = plt.subplots()
    elif fig is None:
      fig = plt.gcf()
    elif ax is None:
      ax = plt.gca()
    self.fig = fig
    self.ax = ax
    self._show_lines = show_lines
    self._lines = []
    self._points_x = None
    self._points_y = None
    self._threshold = threshold
    self._scale = scale
    self._cmap = cmap
    self._scale_alpha_by_coupling = scale_alpha_by_coupling
    self._alpha = alpha
    self._title = title
    self._xlim = xlim
    self._ylim = ylim

  def _scatter(self, ot: Transport):
    """Compute the position and scales of the points on a 2D plot."""
    if not isinstance(ot.geom, pointcloud.PointCloud):
      raise ValueError("So far we only plot PointCloud geometry.")

    x, y = ot.geom.x, ot.geom.y
    a, b = ot.a, ot.b
    x, y = bidimensional(x, y)
    scales_x = a * self._scale * a.shape[0]
    scales_y = b * self._scale * b.shape[0]
    return x, y, scales_x, scales_y

  def _mapping(self, x: jax.Array, y: jax.Array, matrix: jax.Array):
    """Compute the lines representing the mapping between the 2 point clouds."""
    # Only plot the lines with a cost above the threshold.
    u, v = jnp.where(matrix > self._threshold)
    c = matrix[jnp.where(matrix > self._threshold)]
    xy = jnp.concatenate([x[u], y[v]], axis=-1)

    # Check if we want to adjust transparency.
    scale_alpha_by_coupling = self._scale_alpha_by_coupling

    # We can only adjust transparency if max(c) != min(c).
    if scale_alpha_by_coupling:
      min_matrix, max_matrix = jnp.min(c), jnp.max(c)
      scale_alpha_by_coupling = max_matrix != min_matrix

    result = []
    for i in range(xy.shape[0]):
      strength = jnp.max(jnp.array(matrix.shape)) * c[i]
      if scale_alpha_by_coupling:
        normalized_strength = (c[i] - min_matrix) / (max_matrix - min_matrix)
        alpha = self._alpha * float(normalized_strength)
      else:
        alpha = self._alpha

      # Matplotlib's transparency is sensitive to numerical errors.
      alpha = np.clip(alpha, 0.0, 1.0)

      start, end = xy[i, [0, 2]], xy[i, [1, 3]]
      result.append((start, end, strength, alpha))

    return result

  def __call__(self, ot: Transport) -> List[plt.Artist]:
    """Plot couplings in 2-D, using PCA if data is higher dimensional."""
    x, y, sx, sy = self._scatter(ot)
    self._points_x = self.ax.scatter(
        *x.T, s=sx, edgecolors="k", marker="o", label="x"
    )
    self._points_y = self.ax.scatter(
        *y.T, s=sy, edgecolors="k", marker="X", label="y"
    )
    self.ax.legend(fontsize=15)

    if self._title is not None:
      self.ax.set_title(self._title)

    if not self._show_lines:
      return []

    lines = self._mapping(x, y, ot.matrix)
    cmap = plt.get_cmap(self._cmap)
    self._lines = []
    for start, end, strength, alpha in lines:
      line, = self.ax.plot(
          start,
          end,
          linewidth=0.5 + 4 * strength,
          color=cmap(strength),
          zorder=0,
          alpha=alpha
      )
      self._lines.append(line)

    if self._xlim is not None:
      self.ax.set_xlim(self._xlim)
    if self._ylim is not None:
      self.ax.set_ylim(self._ylim)

    return [self._points_x, self._points_y] + self._lines

  def update(self,
             ot: Transport,
             title: Optional[str] = None) -> List[plt.Artist]:
    """Update a plot with a transport instance."""
    x, y, _, _ = self._scatter(ot)
    self._points_x.set_offsets(x)
    self._points_y.set_offsets(y)

    if title is not None:
      self.ax.set_title(title)

    if not self._show_lines:
      return []

    new_lines = self._mapping(x, y, ot.matrix)
    cmap = plt.get_cmap(self._cmap)
    for line, new_line in zip(self._lines, new_lines):
      start, end, strength, alpha = new_line

      line.set_data(start, end)
      line.set_linewidth(0.5 + 4 * strength)
      line.set_color(cmap(strength))
      line.set_alpha(alpha)

    # Maybe add new lines to the plot.
    num_lines = len(self._lines)
    num_to_plot = len(new_lines) if self._show_lines else 0
    for i in range(num_lines, num_to_plot):
      start, end, strength, alpha = new_lines[i]

      line, = self.ax.plot(
          start,
          end,
          linewidth=0.5 + 4 * strength,
          color=cmap(strength),
          zorder=0,
          alpha=alpha
      )
      self._lines.append(line)

    self._lines = self._lines[:num_to_plot]  # Maybe remove some
    return [self._points_x, self._points_y] + self._lines

  def animate(
      self,
      transports: Sequence[Transport],
      titles: Optional[Sequence[str]] = None,
      frame_rate: float = 10.0
  ) -> animation.FuncAnimation:
    """Make an animation from several transports."""
    _ = self(transports[0])
    if titles is None:
      titles = [None for _ in np.arange(0, len(transports))]
    assert len(titles) == len(transports), (
        f"titles/transports lengths differ `{len(titles)}`/`{len(transports)}`."
    )
    return animation.FuncAnimation(
        self.fig,
        lambda i: self.update(transports[i], titles[i]),
        np.arange(0, len(transports)),
        init_func=lambda: self.update(transports[0], titles[0]),
        interval=1000 / frame_rate,
        blit=True
    )


class PlotMM(Plot):
  """Plots an optimal transport map for :class:`~ott.experimental.mmsinkhorn.MMSinkhorn`.

  It enables to either plot or update a plot in a single object, offering the
  possibilities to create animations as a
  :class:`~matplotlib.animation.FuncAnimation`, which can in turned be saved to
  disk at will. There are two design principles here:

  #. we do not rely on saving to/loading from disk to create animations
  #. we try as much as possible to disentangle the transport problem from
       its visualization.

  Args:
    fig: Specify figure object. Created by default
    ax: Specify axes objects. Created by default
    fix_axes_lim: Whether to fix x/y limits to :math:`[0, 1]`.
    cmap: color map used to plot line colors.
    markers: Markers for each marginal.
    alpha: default alpha value for lines.
    title: title of the plot.
  """  # noqa: E501

  def __init__(
      self,
      fig: Optional[plt.Figure] = None,
      ax: Optional[plt.Axes] = None,
      fix_axes_lim: bool = False,
      cmap: Union[str, mcolors.Colormap] = "cividis_r",
      markers: str = "svopxdh",
      alpha: float = 0.6,
      title: Optional[str] = None,
  ):
    if isinstance(cmap, str):
      cmap = plt.colormaps[cmap]
    super().__init__(fig=fig, ax=ax, cmap=cmap, alpha=alpha, title=title)
    self._patches = None
    self._points = None
    self._markers = markers
    self._fix_axes_lim = fix_axes_lim

  def __call__(
      self,
      ot: mmsinkhorn.MMSinkhornOutput,
      top_k: Optional[int] = None
  ) -> List["plt.Artist"]:
    """Plot 2-D couplings. does not support higher dimensional."""
    assert ot.n_marginals <= len(self._markers), "Not enough markers to plot."
    self._points = []
    self._patches = []
    n0 = max(ot.shape)
    top_k = n0 if top_k is None else top_k

    # extract the `top_k` entries in the tensor, and their indices.
    _, idx = jax.lax.top_k(ot.tensor.ravel(), top_k)
    indices = jnp.unravel_index(idx, ot.shape)

    alphas = np.linspace(self._alpha, 0.2, max(0, top_k - n0))
    for j in range(top_k):
      points = [x[indices[i][j], ...] for i, x in enumerate(ot.x_s)]
      # re-order to ensure polygons have maximal area
      points = [points[i] for i in ccworder(jnp.array(points))]
      alpha = self._alpha if j < n0 else alphas[j - n0]

      polygon = ptc.Polygon(
          points,
          fill=True,
          linewidth=2,
          color=self._cmap(float(j >= n0)),
          alpha=alpha,
          zorder=-j,
      )
      self._patches.append(self.ax.add_patch(polygon))

    for i, (x, a) in enumerate(zip(ot.x_s, ot.a_s)):
      points = self.ax.scatter(
          x[:, 0],
          x[:, 1],
          s=200.0 * len(a) * a,
          marker=self._markers[i],
          c="black",
          linewidth=0.0,
          edgecolor=None,
          label=str(i)
      )
      self._points.append(points)

    if self._title is not None:
      self.ax.set_title(self._title)

    return self._points + self._patches

  def update(
      self,
      ot: mmsinkhorn.MMSinkhornOutput,
      title: Optional[str] = None,
      top_k: Optional[int] = None,
  ) -> List[plt.Artist]:
    """Update a plot with a transport instance."""
    n0 = max(ot.shape)
    top_k = n0 if top_k is None else top_k
    # extract the `top_k` entries in the tensor, and their indices.
    _, idx = jax.lax.top_k(ot.tensor.ravel(), top_k)
    indices = jnp.unravel_index(idx, ot.shape)

    alphas = np.linspace(self._alpha, 0.2, max(0, top_k - n0))
    for j, patch in enumerate(self._patches):
      points = [x[indices[i][j], ...] for i, x in enumerate(ot.x_s)]
      # re-order to ensure polygons have maximal area
      points = [points[i] for i in ccworder(jnp.array(points))]
      alpha = self._alpha if j < n0 else alphas[j - n0]
      # update the location of the patches according to the new coordinates
      patch.set_xy(points)
      patch.set_color(self._cmap(float(j >= n0)))
      patch.set_alpha(alpha)

    for points, xs in zip(self._points, ot.x_s):
      points.set_offsets(xs)

    if title is not None:
      self.ax.set_title(title)

    # we keep the axis fixed to 0-1 assuming normalized data
    if self._fix_axes_lim:
      eps = 2.5e-2
      self.ax.set_ylim(-eps, 1.0 + eps)
      self.ax.set_xlim(-eps, 1.0 + eps)

    return self._points + self._patches

  def animate(
      self,
      transports: Sequence[mmsinkhorn.MMSinkhornOutput],
      titles: Optional[Sequence[str]] = None,
      frame_rate: float = 10.0,
      top_k: Optional[int] = None,
  ) -> animation.FuncAnimation:
    """Make an animation from several transports."""
    ot, *_ = transports
    _ = self(ot, top_k=top_k)
    titles = titles if titles is not None else [""] * len(transports)
    return animation.FuncAnimation(
        self.fig,
        lambda i: self.update(ot=transports[i], title=titles[i], top_k=top_k),
        np.arange(0, len(transports)),
        init_func=lambda: self.update(ot, title=titles[0], top_k=top_k),
        interval=1000.0 / frame_rate,
        blit=True,
    )


def get_plotkwargs(
    background: bool,
    *,
    small_alpha: float = 0.2,
    large_alpha: float = 0.7,
    darkmode: bool = False,
    small_size: int = 50,
    mid_size: int = 60,
    size_multiplier: float = 1.2
) -> Dict[str, Any]:
  r"""Generate marker styling specifications for transport visualization.

  This utility function creates a dictionary of matplotlib styling parameters
  for various types of points and arrows used in optimal transport
  visualizations.

  Args:
    background: Whether source and target points should have small alphas to
      de-emphasize them and highlight other elements like dynamic points
      or arrows.
    small_alpha: Transparency value for background points.
    large_alpha: Transparency value for foreground/highlighted points.
    darkmode: Whether to use colors suitable for dark background plots.
      If :obj:`True``, use lighter colors, otherwise use standard colors.
    small_size: Base marker size for regular source/target points.
    mid_size: Marker size for highlighted new source points.
    size_multiplier: Multiplicative factor to enlarge transported points
      relative to their base size.

  Returns:
    A dictionary with the following keys, each containing marker styling
    parameters for matplotlib scatter/quiver plots:

    - ``'x'``: Regular source points :math:`\mu_0`
    - ``'tx'``: Transported source points :math:`\mu_t`
    - ``'xnew'``: New batch of highlighted source points
    - ``'txnew_interm'``: Intermediate positions of new transported points
    - ``'txnew'``: Final positions of new transported points
    - ``'y'``: Target points :math:`\mu_1`
    - ``'ifm'``: Independent flow matching (IFM) interpolated points
    - ``'arrows_grid'``: Velocity field arrows on grid points
    - ``'arrows_dynamic'``: Velocity field arrows for moving points
    - ``'arrows_ifm'``: Velocity field arrows for IFM points
  """
  sourcecolor = "lightcoral" if darkmode else "red"
  newsourcecolor = "salmon" if darkmode else "red"
  targetcolor = "deepskyblue" if darkmode else "blue"
  edgecolor = "white" if darkmode else "black"
  ifmcolor = "palegreen" if darkmode else "green"
  arrowscolor = "white" if darkmode else "black"
  arrows_ifm_color = "palegreen" if darkmode else "green"

  mid_alpha = (large_alpha + small_alpha) / 2
  # Regular points from source
  x = {
      "s": small_size,
      "label": r"$\mu_0$",
      "marker": "o",
      "color": sourcecolor,
      "edgecolor": edgecolor,
      "alpha": small_alpha if background else large_alpha,
  }

  # Points being transported
  tx = {
      "s": small_size * size_multiplier,
      "label": r"$\mu_t$",
      "marker": "o",
      "color": sourcecolor,
      "edgecolor": edgecolor,
      "alpha": large_alpha
  }

  # New batch of source points supposed to be highlighted
  xnew = {
      "s": mid_size,
      "marker": "h",
      "edgecolor": edgecolor,
      "color": newsourcecolor,
      "alpha": large_alpha,
  }

  txnew_interm = {
      "s": mid_size * size_multiplier,
      "marker": "h",
      "edgecolor": edgecolor,
      "color": newsourcecolor,
      "alpha": small_alpha,
  }

  txnew = {
      "s": mid_size * size_multiplier,
      "marker": "h",
      "edgecolor": edgecolor,
      "color": newsourcecolor,
      "alpha": large_alpha,
  }

  # Target Points
  y = {
      "s": small_size,
      "label": r"$\mu_1$",
      "marker": "s",
      "edgecolor": edgecolor,
      "color": targetcolor,
      "alpha": small_alpha if background else large_alpha,
  }

  # IFM Points
  ifm = {
      "s": small_size,
      "label": r"IFM $\mu_t$",
      "marker": "d",
      "edgecolor": edgecolor,
      "color": ifmcolor,
      "alpha": mid_alpha,
  }

  arrows_ifm = {"color": arrows_ifm_color, "alpha": large_alpha}

  arrows_grid = {
      "color": arrowscolor,
      "alpha": mid_alpha if background else large_alpha
  }

  arrows_dynamic = {"color": arrowscolor, "alpha": mid_alpha}

  return {
      "x": x,
      "tx": tx,
      "xnew": xnew,
      "txnew_interm": txnew_interm,
      "txnew": txnew,
      "y": y,
      "ifm": ifm,
      "arrows_grid": arrows_grid,
      "arrows_dynamic": arrows_dynamic,
      "arrows_ifm": arrows_ifm,
  }


def transport_animation(
    n_frames: int,
    static_src_points: jax.Array,
    static_tgt_points: jax.Array,
    *,
    n_grid: int = 0,
    velocity_field: Optional[Callable[[jax.Array, jax.Array],
                                      jax.Array]] = None,
    dynamic_src_points: Optional[jax.Array] = None,
    num_ifm_interpolants: Union[int, Literal["all"]] = 0,
    plot_ifm_arrows: bool = False,
    title: Optional[str] = None,
    figsize: Tuple[int, int] = (8, 6),
    xlimits: Optional[Tuple[float, float]] = None,
    ylimits: Optional[Tuple[float, float]] = None,
    padding: float = 0.1,
    interval: int = 300,
    save_path: Optional[str] = None,
) -> Union[plt.Figure, animation.FuncAnimation]:
  r"""Create animated visualizations of optimal transport and flow matching.

  This function generates animations illustrating various aspects of optimal
  transport, including :term`Monge map`, McCann interpolation,
  :term:`Benamou-Brenier` velocity fields, and flow matching approaches.
  It supports multiple visualization modes and can display static point clouds,
  dynamic trajectories, and velocity fields on grids.

  Args:
    n_frames: Number of animation frames :math:`> 1`. If ``1``,
      creates a static plot instead of an animation.
    static_src_points: Source distribution points of shape ``[n, 2]``,
      representing samples from :math:`\mu_0`.
    static_tgt_points: Target distribution points of shape ``[n, 2]``,
      representing samples from :math:`\mu_1`.
    n_grid: Number of grid points per dimension for displaying velocity fields.
      If :math:`> 0`, displays velocity field arrows on a uniform
      :math:`n_{grid} \times n_{grid}` grid.
    velocity_field: Optional learned/estimated velocity field function with a
      signature ``v(t, x) -> velocity`` where ``t`` is time of shape
      ``[batch,]`` and ``x`` is position of shape ``[batch, 2]``.
    dynamic_src_points: Additional points of shape ``[m, 2]`` to highlight and
      transport dynamically through the animation. Useful for emphasizing
      specific trajectories. If :obj:`None` and the velocity field is specified,
      use ``static_src_points``.
    num_ifm_interpolants: Number of random pairs of source and target points for
      the independent flow matching (IFM) interpolant :math:`(1-t)x_0 + tx_1`.
      If ``'all'``, use ``len(static_src_points) ** 2``.
    plot_ifm_arrows: Whether to display velocity arrows for the IFM
      interpolant. Only relevant when ``num_ifm_interpolants > 0``.
    title: Title for the animation/plot.
    figsize: Figure size as ``(width, height)`` in inches.
    xlimits: X-axis limits as ``(xmin, xmax)``. If :obj:`None`, it is computed
      automatically from all points with padding.
    ylimits: Y-axis limits as ``(ymin, ymax)``. If :obj:`None`, it is computed
      automatically from all points with padding.
    padding: Fractional padding to add around automatically computed axis
      limits. For example, ``0.1`` adds 10% padding on each side.
    interval: Used for animations, delay between frames in milliseconds.
    save_path: Path where to save the animation/plot.

  Returns:
    An animation object if ``n_frames > 1``, otherwise a figure.
  """
  assert n_frames >= 1, f"n_frames must be > 0, got {n_frames}."
  assert static_tgt_points.shape == static_src_points.shape
  if n_grid > 0:
    assert velocity_field is not None, \
      "To display field on grid points, please provide the velocity fields."

  n_static = len(static_src_points)
  if num_ifm_interpolants == "all":
    num_ifm_interpolants = n_static ** 2
  if velocity_field is not None and dynamic_src_points is None:
    dynamic_src_points = static_src_points
  n_dynamic = 0 if dynamic_src_points is None else len(dynamic_src_points)

  plot_transport = velocity_field is not None and dynamic_src_points is not None

  # If we are plotting extra stuff on top of data,
  # data is displayed in low alpha as background
  background = plot_ifm_arrows or plot_transport
  dict_pk = get_plotkwargs(background=background)

  # Time parameterization
  times = jnp.linspace(0.0, 1.0, n_frames)
  delta_times = jnp.diff(times)
  t0 = times[0]
  # scale of arrows for first step (and maybe last-and-only step).
  dt0 = 1.0 if n_frames == 1 else delta_times[0]

  fig, ax = plt.subplots(figsize=figsize)
  fig.tight_layout(pad=2.0)
  ax.scatter(
      static_src_points[:, 0],
      static_src_points[:, 1],
      **dict_pk["x"],
  )
  ax.scatter(
      static_tgt_points[:, 0],
      static_tgt_points[:, 1],
      **dict_pk["y"],
  )

  # Define space of points that will move (all by default if arrows
  # are requested and no dynamic_points are passed)
  if plot_transport:
    v = velocity_field(jnp.full(n_dynamic, fill_value=t0), dynamic_src_points)
    quiver_points = ax.quiver(
        dynamic_src_points[:, 0],
        dynamic_src_points[:, 1],
        dt0 * v[:, 0],
        dt0 * v[:, 1],
        angles="xy",
        scale_units="xy",
        scale=1,
        **dict_pk["arrows_dynamic"],
    )
    # Add dynamic points
    scatter_interm_points_before = ax.scatter(
        dynamic_src_points[:, 0],
        dynamic_src_points[:, 1],
        **dict_pk["txnew_interm"],
    )
    if n_frames > 1:
      scatter_interm_points_after = ax.scatter(
          dynamic_src_points[:, 0] + dt0 * v[:, 0],
          dynamic_src_points[:, 1] + dt0 * v[:, 1],
          **dict_pk["txnew"],
      )

  # Gather all points to set limits adaptively if needed.
  all_points = [static_src_points, static_tgt_points]
  if plot_transport:
    all_points.append(dynamic_src_points)
  all_points = jnp.concatenate(all_points, axis=0)
  if xlimits is None:
    xlimits = jnp.min(all_points[:, 0]), jnp.max(all_points[:, 0])
    xscale = xlimits[1] - xlimits[0]
    xlimits = (xlimits[0] - padding * xscale, xlimits[1] + padding * xscale)
  if ylimits is None:
    ylimits = jnp.min(all_points[:, 1]), jnp.max(all_points[:, 1])
    yscale = ylimits[1] - ylimits[0]
    ylimits = (ylimits[0] - padding * yscale, ylimits[1] + padding * yscale)
  del all_points

  # Display velocities on grids.
  if n_grid > 0:
    x = jnp.linspace(*xlimits, n_grid)
    y = jnp.linspace(*ylimits, n_grid)
    X, Y = jnp.meshgrid(x, y)
    points_grid = jnp.stack([X, Y], axis=-1).reshape(-1, 2)

    t = jnp.full(len(points_grid), fill_value=t0)
    v_grid = velocity_field(t, points_grid)

    quiver_grid = ax.quiver(
        points_grid[:, 0],
        points_grid[:, 1],
        dt0 * v_grid[:, 0],
        dt0 * v_grid[:, 1],
        angles="xy",
        scale_units="xy",
        scale=1,
        **dict_pk["arrows_grid"],
    )

  if num_ifm_interpolants:
    product_points = jnp.stack([
        jnp.repeat(static_src_points, axis=0, repeats=n_static),
        jnp.tile(static_tgt_points, reps=(n_static, 1)),
    ])

    if num_ifm_interpolants < n_static ** 2:
      subset_ixs = jr.choice(
          jr.key(0), n_static ** 2, (num_ifm_interpolants,), replace=False
      )
      product_points = product_points[:, subset_ixs, :]

    product_points_at_t = product_points[0, :, :]
    prod_scatter = ax.scatter(
        product_points_at_t[:, 0],
        product_points_at_t[:, 1],
        **dict_pk["ifm"],
    )
    if plot_ifm_arrows:
      prod_quiver = ax.quiver(
          product_points_at_t[:, 0],
          product_points_at_t[:, 1],
          dt0 * (product_points[1, :, 0] - product_points[0, :, 0]),
          dt0 * (product_points[1, :, 1] - product_points[0, :, 1]),
          angles="xy",
          scale_units="xy",
          scale=1,
          **dict_pk["arrows_ifm"],
      )

  ax.set_title(title)
  ax.legend()
  ax.grid(True)
  ax.set_aspect("equal")
  ax.set_xlim(*xlimits)
  ax.set_ylim(*ylimits)
  # End of static frame

  # Initialize dynamic points at time `t`
  dyn_points_t = dynamic_src_points

  def update_frame(frame: int) -> None:
    nonlocal dyn_points_t
    t = times[frame]
    dt = delta_times[frame - 1] if frame > 0 else dt0
    # Update grid arrows (locations stay fixed)
    if n_grid > 0 and t < 1.0:
      times_t = jnp.full((points_grid.shape[0]), fill_value=t)
      v_grid = velocity_field(times_t, points_grid)
      quiver_grid.set_UVC(dt * v_grid[:, 0], dt * v_grid[:, 1])

    # Update moving point arrows (locations move with time)
    if plot_transport:
      if t >= 1.0:
        v_points = jnp.zeros_like(dyn_points_t)
      else:
        v_points = velocity_field(
            jnp.full(n_dynamic, fill_value=t), dyn_points_t
        )

      quiver_points.set_offsets(dyn_points_t)
      quiver_points.set_UVC(dt * v_points[:, 0], dt * v_points[:, 1])
      scatter_interm_points_before.set_offsets(dyn_points_t)
      scatter_interm_points_after.set_offsets(dyn_points_t + v_points * dt)

      dyn_points_t = dyn_points_t + dt * v_points

    if n_grid > 0 or plot_transport:
      ax.set_title(f"{title} at time {t:.2f}")

    if num_ifm_interpolants:
      pp_at_t = (1 - t) * product_points[0, :, :] + t * product_points[1, :, :]
      prod_scatter.set_offsets(pp_at_t)

      if plot_ifm_arrows:
        prod_quiver.set_offsets(pp_at_t)
      ax.set_title(f"{title} at time {t:.2f}")

  if n_frames == 1:
    if save_path is not None:
      plt.savefig(save_path)
    return fig

  ani = animation.FuncAnimation(
      fig,
      update_frame,
      frames=n_frames,
      init_func=lambda: None,
      blit=False,
      interval=interval,
      repeat=n_frames > 1,
  )
  if save_path is not None:
    ani.save(save_path, bitrate=2000)
  plt.close()
  return ani


@jax.jit
def ccworder(A: jax.Array) -> jax.Array:
  """Order points in counter-clockwise direction for polygon plotting.

  This helper function reorders a set of 2D points so that they can be used to
  draw a polygon with maximal area. It centers the points at the origin and
  then sorts them by their angular position.

  Args:
    A: Array of shape ``[n, 2]`` containing 2D point coordinates.

  Returns:
    Array of indices that reorder the input points in counter-clockwise order
    starting from the angle 0 (positive x-axis).

  Note:
    Based on: https://stackoverflow.com/questions/5040412/how-to-draw-the-largest-polygon-from-a-set-of-points
  """
  A = A - jnp.mean(A, 0, keepdims=True)
  return jnp.argsort(jnp.arctan2(A[:, 1], A[:, 0]))


def bidimensional(x: jax.Array, y: jax.Array) -> Tuple[jax.Array, jax.Array]:
  """Apply PCA to reduce to bi-dimensional data."""
  if x.shape[1] < 3:
    return x, y

  u, s, _ = sp.linalg.svds(np.array(jnp.concatenate([x, y], axis=0)), k=2)
  proj = u * s
  k = x.shape[0]
  return proj[:k], proj[k:]



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Callable, Dict, Literal, Optional, Sequence, Tuple

import jax
import jax.numpy as jnp

from ott.geometry import geometry, pointcloud
from ott.math import utils as math_utils

__all__ = ["LinearProblem"]

# TODO(michalk8): move to typing.py when refactoring the types
MarginalFunc = Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray]
TransportAppFunc = Callable[[jnp.ndarray, jnp.ndarray, jnp.ndarray, int],
                            jnp.ndarray]


@jax.tree_util.register_pytree_node_class
class LinearProblem:
  r"""Linear OT problem.

  This class describes the main ingredients appearing in a linear OT problem.
  Namely, a ``geom`` object (including cost structure/points) describing point
  clouds or the support of measures, followed by probability masses ``a`` and
  ``b``. Unbalancedness of the problem is also kept track of, through two
  coefficients ``tau_a`` and ``tau_b``, which are both kept between 0 and 1
  (1 corresponding to a balanced OT problem).

  Args:
    geom: The ground geometry cost of the linear problem.
    a: The first marginal. If :obj:`None`, it will be uniform.
    b: The second marginal. If :obj:`None`, it will be uniform.
    tau_a: If :math:`<1`, defines how much unbalanced the problem is
      on the first marginal.
    tau_b: If :math:`< 1`, defines how much unbalanced the problem is
      on the second marginal.
  """

  def __init__(
      self,
      geom: geometry.Geometry,
      a: Optional[jnp.ndarray] = None,
      b: Optional[jnp.ndarray] = None,
      tau_a: float = 1.0,
      tau_b: float = 1.0
  ):
    self.geom = geom
    self._a = a
    self._b = b
    self.tau_a = tau_a
    self.tau_b = tau_b

  @property
  def a(self) -> jnp.ndarray:
    """First marginal."""
    if self._a is not None:
      return self._a
    n, _ = self.geom.shape
    return jnp.full((n,), fill_value=1.0 / n, dtype=self.dtype)

  @property
  def b(self) -> jnp.ndarray:
    """Second marginal."""
    if self._b is not None:
      return self._b
    _, m = self.geom.shape
    return jnp.full((m,), fill_value=1.0 / m, dtype=self.dtype)

  @property
  def is_balanced(self) -> bool:
    """Whether the problem is balanced."""
    return self.tau_a == 1.0 and self.tau_b == 1.0

  @property
  def is_uniform(self) -> bool:
    """True if no weights ``a,b`` were passed, and have defaulted to uniform."""
    return self._a is None and self._b is None

  @property
  def is_equal_size(self) -> bool:
    """True if square shape, i.e. ``n == m``."""
    return self.geom.shape[0] == self.geom.shape[1]

  @property
  def is_assignment(self) -> bool:
    """True if assignment problem."""
    return self.is_equal_size and self.is_uniform and self.is_balanced

  @property
  def epsilon(self) -> float:
    """Entropic regularization."""
    return self.geom.epsilon

  @property
  def dtype(self) -> jnp.dtype:
    """The data type of the geometry."""
    return self.geom.dtype

  def potential_fn_from_dual_vec(
      self,
      fg: jax.Array,
      *,
      epsilon: Optional[float] = None,
      axis: Literal[0, 1],
  ) -> Callable[[jax.Array], jax.Array]:
    r"""Get potential function from a dual vector using the :term:`c-transform`.

    Args:
      fg: Potential vector :math:`\mathbb{f}` if ``axis = 0``
        else :math:`\mathbb{g}` of shape ``[n,]`` or ``[m,]``, respectively.
      epsilon: Epsilon regularization. If :obj:`None`, use in the :attr:`geom`.
      axis: If ``axis = 0``, return the :math:`g`-potential function, otherwise
        return the :math:`f`-potential function.

    Returns:
      The dual potential function.
    """

    def f_potential(x: jax.Array) -> jax.Array:
      x, y = jnp.atleast_2d(x), self.geom.y
      geom = pointcloud.PointCloud(x, y, cost_fn=self.geom.cost_fn)
      prob = LinearProblem(geom, b=self.b)
      f, _ = prob._c_transform(fg, epsilon=epsilon, axis=axis)
      return f.squeeze(0)

    def g_potential(y: jax.Array) -> jax.Array:
      x, y = self.geom.x, jnp.atleast_2d(y)
      geom = pointcloud.PointCloud(x, y, cost_fn=self.geom.cost_fn)
      prob = LinearProblem(geom, a=self.a)
      g, _ = prob._c_transform(fg, epsilon=epsilon, axis=axis)
      return g.squeeze(0)

    assert axis in (0, 1), axis
    epsilon = self.geom.epsilon if epsilon is None else epsilon
    return g_potential if axis == 0 else f_potential

  def _c_transform(
      self,
      fg: jax.Array,
      *,
      epsilon: Optional[float] = None,
      axis: Literal[0, 1],
  ) -> Tuple[jax.Array, jax.Array]:

    def _soft_c_transform(fg: jax.Array) -> Tuple[jax.Array, jax.Array]:
      cost = self.geom.cost_matrix
      z = (fg - cost) / epsilon
      return -epsilon * math_utils.logsumexp(z, b=self.b, axis=axis), z

    def _hard_c_transform(fg: jax.Array) -> Tuple[jax.Array, jax.Array]:
      cost = self.geom.cost_matrix
      z = fg - cost
      pos_weights = self.b[None, :] > 0.0
      return -jnp.max(z, initial=-jnp.inf, where=pos_weights, axis=axis), z

    assert axis in (0, 1), axis
    fg = jnp.expand_dims(fg, 1 - axis)
    epsilon = self.geom.epsilon if epsilon is None else epsilon
    return jax.lax.cond(epsilon > 0.0, _soft_c_transform, _hard_c_transform, fg)

  def get_transport_functions(
      self, lse_mode: bool
  ) -> Tuple[MarginalFunc, MarginalFunc, TransportAppFunc]:
    """Instantiate useful functions for Sinkhorn depending on lse_mode."""
    geom = self.geom
    if lse_mode:
      marginal_a = lambda f, g: geom.marginal_from_potentials(f, g, 1)
      marginal_b = lambda f, g: geom.marginal_from_potentials(f, g, 0)
      app_transport = geom.apply_transport_from_potentials
    else:
      marginal_a = lambda f, g: geom.marginal_from_scalings(
          geom.scaling_from_potential(f), geom.scaling_from_potential(g), 1
      )
      marginal_b = lambda f, g: geom.marginal_from_scalings(
          geom.scaling_from_potential(f), geom.scaling_from_potential(g), 0
      )
      app_transport = lambda f, g, z, axis: geom.apply_transport_from_scalings(
          geom.scaling_from_potential(f), geom.scaling_from_potential(g), z,
          axis
      )
    return marginal_a, marginal_b, app_transport

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:  # noqa: D102
    return ([self.geom, self._a, self._b], {
        "tau_a": self.tau_a,
        "tau_b": self.tau_b
    })

  @classmethod
  def tree_unflatten(  # noqa: D102
      cls, aux_data: Dict[str, Any], children: Sequence[Any]
  ) -> "LinearProblem":
    return cls(*children, **aux_data)



from typing import Any

import jax
import jax.numpy as jnp
import jax.scipy as jsp
import torchvision
import torchvision.transforms as transforms
from torch.utils.data import DataLoader

__all__ = ["get_cifar10"]


def to_fp_array(arr: Any) -> jnp.array:
  dtype = jnp.float64 if jax.config.jax_enable_x64 else jnp.float32
  return jnp.asarray(arr, dtype=dtype)


def get_cifar10(
    *,
    sigma: float = 0.5,
    gaussian_blur_kernel_size: int = 2,
    root: str = "data/cifar10",
    batch_size: int = 1000,
    use_flip: bool = True
) -> tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:

  transform = transforms.Compose([
      transforms.ToTensor(),
      transforms.Normalize((0.5,), (0.5,)),
  ])

  # Download and load datasets
  ds = torchvision.datasets.CIFAR10(
      root=root, train=True, download=True, transform=transform
  )
  loader = DataLoader(ds, batch_size=batch_size, shuffle=False)

  if use_flip:
    transform_flip = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize((0.5,), (0.5,)),
        transforms.RandomHorizontalFlip(p=1.0)
    ])

    ds_flip = torchvision.datasets.CIFAR10(
        root=root, train=True, download=True, transform=transform_flip
    )
    loader_flip = DataLoader(ds_flip, batch_size=batch_size, shuffle=False)
  else:
    loader_flip = []

  src_batches = []
  target_batches = []
  label_batches = []

  def psd_gaussian_blur(x: jnp.ndarray) -> jnp.ndarray:
    assert x.ndim == 4, x.shape
    n, c, h, w = x.shape
    assert h == w
    # 1D Gaussian kernel
    ys = jnp.arange(-gaussian_blur_kernel_size, gaussian_blur_kernel_size + 1)
    kernel_1d = jsp.stats.norm.pdf(ys, scale=sigma)
    kernel_1d = kernel_1d / jnp.sum(kernel_1d)  # Normalize

    # Flatten input to apply convolution
    x = x.reshape(n * c, h, w)

    def apply_separable_conv(img):
      # Convolve along rows
      img_h = jsp.signal.convolve(img, kernel_1d[None, :], mode="same")
      # Convolve along columns
      return jsp.signal.convolve(img_h, kernel_1d[:, None], mode="same")

    conv = jax.vmap(apply_separable_conv)
    blurred = conv(x)

    return blurred.reshape(n, c, h, w)

  for ld in [loader, loader_flip] if use_flip else [loader]:
    for batch_src, batch_label in ld:
      batch_src = jnp.array(batch_src.numpy())

      batch_target = psd_gaussian_blur(batch_src)

      src_batches.append(batch_src.reshape(batch_src.shape[0], -1))
      target_batches.append(batch_target.reshape(batch_target.shape[0], -1))
      label_batches.append(jnp.array(batch_label.numpy()))

  src = jnp.concatenate(src_batches, axis=0)
  target = jnp.concatenate(target_batches, axis=0)
  labels = jnp.concatenate(label_batches, axis=0)

  return to_fp_array(src), to_fp_array(target), labels



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from typing import Any, Callable

import jax
import jax.numpy as jnp

from ott import math

__all__ = ["velocity_from_brenier_potential"]


def velocity_from_brenier_potential(
    potential: Callable[[jnp.ndarray], jnp.ndarray],
    **kwargs: Any,
) -> Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray]:
  """Get optimal time-dependent velocity field from :term:`Brenier potential`.

  The solution is computed numerically using a :term:`Legendre transform`.

  Args:
    potential: A convex potential of shape ``[d,]``.
    kwargs: Keyword arguments for :func:`~ott.math.legendre`.

  Returns:
    A time-parameterized velocity field ``vel(t, x)`` that expects time array
    of shape ``[n,]`` and inputs of shape ``[n, d]``.
  """

  @functools.partial(jax.vmap, in_axes=[0, 0])
  def vel(t: jnp.array, z: jnp.array) -> jnp.array:

    def pot_t(x: jnp.ndarray) -> jnp.ndarray:
      return 0.5 * (1 - t) * jnp.sum(x ** 2) + t * potential(x)

    grad_pot_t_star = jax.grad(math.legendre(pot_t, **kwargs))
    x = jax.lax.cond(t == 0.0, lambda z: z, grad_pot_t_star, z)
    return jax.grad(potential)(x) - x

  return vel



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Callable, Literal, Optional, Tuple, Union

import jax
import jax.numpy as jnp
import jax.random as jr
import jax.tree_util as jtu

from ott.geometry import costs, pointcloud

__all__ = ["SemidiscretePointCloud"]


@jtu.register_pytree_node_class
class SemidiscretePointCloud:
  """Semidiscrete point cloud geometry.

  Instances of this geometry can be sampled using the :meth:`sample` method.

  Args:
    sampler: Function with a signature ``(rng, shape, dtype) -> array``
      corresponding to the source distribution.
    y: Array of shape ``[m, ...]`` corresponding to the target distribution.
    cost_fn: Cost function. If :obj:`None`,
      use :class:`~ott.geometry.costs.SqEuclidean`.
    epsilon: Regularization parameter. Can be set to :math:`0` to solve the
      unregularized :term:`semidiscrete problem`.
    relative_epsilon: Whether ``epsilon`` refers to a fraction of the
      :attr:`~ott.geometry.pointcloud.PointCloud.mean_cost_matrix` or
      :attr:`~ott.geometry.pointcloud.PointCloud.std_cost_matrix`.
    scale_cost: Option to rescale the cost matrix.
    relative_epsilon_seed: Random seed when estimating the :attr:`epsilon`.
    relative_epsilon_num_samples: Number of samples when estimating
      the :attr:`epsilon`.
  """

  def __init__(
      self,
      sampler: Callable[[jax.Array, Tuple[int, ...], Optional[jnp.dtype]],
                        jax.Array],
      y: jax.Array,
      cost_fn: Optional[costs.CostFn] = None,
      epsilon: Optional[Union[float, jax.Array]] = None,
      relative_epsilon: Optional[Literal["mean", "std"]] = None,
      scale_cost: Union[float, Literal["mean", "max_norm", "max_bound",
                                       "max_cost", "median"]] = 1.0,
      relative_epsilon_seed: int = 0,
      relative_epsilon_num_samples: int = 1024,
  ):
    assert relative_epsilon_num_samples > 0, \
      "Number of samples when estimating relative epsilon must be positive."
    self.sampler = sampler
    self.y = y
    self.cost_fn = costs.SqEuclidean() if cost_fn is None else cost_fn
    self._epsilon = epsilon
    self._relative_epsilon = relative_epsilon
    self._scale_cost = scale_cost
    self._relative_epsilon_seed = relative_epsilon_seed
    self._relative_epsilon_num_samples = relative_epsilon_num_samples

  def sample(
      self,
      rng: jax.Array,
      num_samples: int,
      *,
      epsilon: Optional[float] = None
  ) -> pointcloud.PointCloud:
    """Sample a point cloud.

    .. note::
      When :attr:`is_entropy_regularized = False <is_entropy_regularized>`,
      some methods and attributes of the sampled
      :class:`~ott.geometry.pointcloud.PointCloud` are not meaningful.
      However, this does not impact the usage of the
      :class:`~ott.solvers.linear.semidiscrete.SemidiscreteSolver`.

    Args:
      rng: Random key used for seeding.
      num_samples: Number of samples.
      epsilon: Epsilon regularization. If :obj:`None`, use :attr:`epsilon`.

    Returns:
      The sampled point cloud.
    """
    assert num_samples > 0, f"Number of samples must be > 0, got {num_samples}."
    shape = (num_samples, *self.y.shape[1:])
    x = self.sampler(rng, shape, self.dtype)
    if epsilon is None:
      epsilon = self.epsilon
    return self._from_samples(x, epsilon)

  def _from_samples(
      self, x: jax.Array, epsilon: jax.Array
  ) -> pointcloud.PointCloud:
    return pointcloud.PointCloud(
        x,
        self.y,
        cost_fn=self.cost_fn,
        epsilon=epsilon,
        relative_epsilon=self._relative_epsilon,
        scale_cost=self._scale_cost,
    )

  @property
  def epsilon(self) -> jax.Array:
    """Epsilon regularization value."""
    if not self.is_entropy_regularized:
      return jnp.array(0.0, dtype=self.dtype)
    rng = jr.key(self._relative_epsilon_seed)
    shape = (self._relative_epsilon_num_samples, *self.y.shape[1:])
    x = self.sampler(rng, shape, self.dtype)
    geom = self._from_samples(x, self._epsilon)
    return jnp.array(geom.epsilon, dtype=self.dtype)

  @property
  def is_entropy_regularized(self) -> bool:
    """Whether ``epsilon > 0``."""
    return self._epsilon is None or self._epsilon > 0.0

  @property
  def shape(self) -> tuple[float, int]:
    """Shape of the geometry."""
    return float("inf"), self.y.shape[0]

  @property
  def dtype(self) -> jnp.dtype:
    """The data type."""
    return self.y.dtype

  def tree_flatten(self):  # noqa: D102
    return (self.y, self.cost_fn), (
        self.sampler, {
            "epsilon": self._epsilon,
            "relative_epsilon": self._relative_epsilon,
            "scale_cost": self._scale_cost,
            "relative_epsilon_seed": self._relative_epsilon_seed,
            "relative_epsilon_num_samples": self._relative_epsilon_num_samples,
        }
    )

  @classmethod
  def tree_unflatten(cls, aux_data, children):  # noqa: D102
    y, cost_fn = children
    sampler, aux_data = aux_data
    return cls(sampler, y, cost_fn=cost_fn, **aux_data)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

import jax.numpy as jnp
import jax.tree_util as jtu

__all__ = ["Epsilon", "DEFAULT_EPSILON_SCALE"]

#: Scaling applied to statistic (mean/std) of cost to compute default epsilon.
DEFAULT_EPSILON_SCALE = 0.05


@jtu.register_pytree_node_class
class Epsilon:
  r"""Scheduler class for the regularization parameter epsilon.

  An epsilon scheduler outputs a regularization strength, to be used by the
  :term:`Sinkhorn algorithm` or variant, at any iteration count. That value is
  either the final, targeted regularization, or one that is larger, obtained by
  geometric decay of an initial multiplier.

  Args:
    target: The epsilon regularizer that is targeted.
    init: Initial value when using epsilon scheduling, understood as a multiple
      of the ``target``, following :math:`\text{init} \text{decay}^{\text{it}}`.
    decay: Geometric decay factor, :math:`\leq 1`.
  """

  def __init__(self, target: jnp.array, init: float = 1.0, decay: float = 1.0):
    assert decay <= 1.0, f"Decay must be <= 1, found {decay}."
    self.target = target
    self.init = init
    self.decay = decay

  def __call__(self, it: Optional[int]) -> jnp.array:
    """Intermediate regularizer value at a given iteration number.

    Args:
      it: Current iteration. If :obj:`None`, return :attr:`target`.

    Returns:
      The epsilon value at the iteration.
    """
    if it is None:
      return self.target
    # the multiple is either 1.0 or a larger init value that is decayed.
    multiple = jnp.maximum(self.init * (self.decay ** it), 1.0)
    return multiple * self.target

  def __repr__(self) -> str:
    return (
        f"{self.__class__.__name__}(target={self.target:.4f}, "
        f"init={self.init:.4f}, decay={self.decay:.4f})"
    )

  def tree_flatten(self):  # noqa: D102
    return (self.target,), {"init": self.init, "decay": self.decay}

  @classmethod
  def tree_unflatten(cls, aux_data, children):  # noqa: D102
    return cls(*children, **aux_data)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import math
from typing import TYPE_CHECKING, Any, Callable, Optional, Tuple, Union

import jax
import jax.experimental.sparse as jesp
import jax.numpy as jnp
import jax.random as jr
import jax.scipy as jsp
import jax.tree_util as jtu

import optax

from ott import utils
from ott.geometry import geometry, pointcloud
from ott.geometry import semidiscrete_pointcloud as sdpc
from ott.math import fixed_point_loop
from ott.problems.linear import linear_problem, potentials
from ott.problems.linear import semidiscrete_linear_problem as sdlp
from ott.solvers.linear import sinkhorn

if TYPE_CHECKING:
  from ott.neural.data import semidiscrete_dataloader

__all__ = [
    "SemidiscreteState",
    "HardAssignmentOutput",
    "SemidiscreteOutput",
    "SemidiscreteSolver",
    "constant_epsilon_scheduler",
]


@jtu.register_dataclass
@dataclasses.dataclass(frozen=True)
class SemidiscreteState:
  """State of the :class:`SemidiscreteSolver`.

  Args:
    it: Iteration number.
    epsilon: Epsilon value at the current iteration.
    g: Dual potential.
    g_ema: Exponential moving average of the dual potential.
    opt_state: State of the optimizer.
    losses: Dual losses.
    grad_norms: Norms of the gradients.
    errors: Marginal deviation errors.
  """
  it: jax.Array
  epsilon: jax.Array
  g: jax.Array
  g_ema: jax.Array
  opt_state: Any
  losses: jax.Array
  grad_norms: jax.Array
  errors: jax.Array


@jtu.register_dataclass
@dataclasses.dataclass(frozen=True)
class HardAssignmentOutput:
  r"""Unregularized linear OT solution.

  Args:
    ot_prob: Linear OT problem.
    paired_indices: Array of shape ``[2, n]``, of :math:`n` pairs
      of indices, for which the optimal transport assigns mass. Namely, for each
      index :math:`0 \leq k < n`, if one has
      :math:`i := \text{paired_indices}[0, k]` and
      :math:`j := \text{paired_indices}[1, k]`, then point :math:`i` in
      the first geometry sends mass to point :math:`j` in the second.
    f: The first dual potential.
    g: The second dual potential.
  """
  ot_prob: linear_problem.LinearProblem
  paired_indices: jax.Array
  f: Optional[jax.Array] = None
  g: Optional[jax.Array] = None

  @property
  def matrix(self) -> jesp.BCOO:
    """Transport matrix of shape ``[n, m]`` with ``n`` non-zero entries."""
    n, m = self.geom.shape
    unit_mass = jnp.full(n, fill_value=1.0 / n, dtype=self.geom.dtype)
    indices = self.paired_indices.T
    return jesp.BCOO((unit_mass, indices), shape=(n, m))

  @property
  def primal_cost(self) -> jax.Array:
    """Transport cost of the linear OT solution."""
    geom = self.geom
    weights = self.matrix.data
    i, j = self.paired_indices[0], self.paired_indices[1]
    if isinstance(geom, pointcloud.PointCloud):
      cost = jax.vmap(geom.cost_fn)(geom.x[i], geom.y[j])
    else:
      cost = geom.cost_matrix[i, j]
    return jnp.sum(weights * cost)

  @property
  def dual_cost(self) -> jax.Array:
    """Dual transport cost."""
    assert self.f is not None, "Dual potential `f` is not computed."
    assert self.g is not None, "Dual potential `g` is not computed."
    return jnp.dot(self.ot_prob.a, self.f) + jnp.dot(self.ot_prob.b, self.g)

  @property
  def geom(self) -> geometry.Geometry:  # noqa: D102
    """Geometry."""
    return self.ot_prob.geom


@jtu.register_dataclass
@dataclasses.dataclass(frozen=True)
class SemidiscreteOutput:
  """Output of the :class:`SemidiscreteSolver`.

  Args:
    g: Dual potential.
    prob: Semidiscrete OT problem.
    it: Final iteration number.
    losses: Dual losses.
    errors: Marginal deviation errors.
    converged: Whether the solver converged.
  """
  g: jax.Array
  prob: sdlp.SemidiscreteLinearProblem
  it: Optional[int] = None
  losses: Optional[jax.Array] = None
  errors: Optional[jax.Array] = None
  converged: Optional[bool] = None

  def sample(
      self,
      rng: jax.Array,
      num_samples: int,
      *,
      epsilon: Optional[float] = None,
  ) -> Union[sinkhorn.SinkhornOutput, HardAssignmentOutput]:
    """Sample a point cloud and compute the OT solution.

    Args:
      rng: Random key used for seeding.
      num_samples: Number of samples.
      epsilon: Epsilon regularization. If :obj:`None`, use the one stored
        in the :attr:`geometry <geom>`.

    Returns:
      The sampled output.
    """
    prob = self.prob.sample(rng, num_samples, epsilon=epsilon)
    is_entreg = self.geom.is_entropy_regularized if epsilon is None else (
        epsilon > 0.0
    )

    if is_entreg:
      f, _ = prob._c_transform(self.g, axis=1)
      # SinkhornOutput's potentials must contain prob. weight normalization
      f_tilde = f + prob.epsilon * jnp.log(1.0 / num_samples)
      g_tilde = self.g + prob.epsilon * jnp.log(prob.b)

      return sinkhorn.SinkhornOutput(
          potentials=(f_tilde, g_tilde),
          ot_prob=prob,
      )

    f, _ = prob._c_transform(self.g, axis=1)
    z = self.g[None, :] - prob.geom.cost_matrix
    row_ixs = jnp.arange(num_samples)
    col_ixs = jnp.argmax(jnp.where(prob.b[None, :], z, -jnp.inf), axis=-1)

    return HardAssignmentOutput(
        prob,
        paired_indices=jnp.stack([row_ixs, col_ixs]),
        f=f,
        g=self.g,
    )

  def to_dual_potentials(
      self, epsilon: Optional[float] = None
  ) -> potentials.DualPotentials:
    """Compute the dual potential function :math:`f`.

    Args:
      epsilon: Epsilon regularization. If :obj:`None`, use the one stored
        in the :attr:`geometry <geom>`.

    Returns:
      The dual potential :math:`f`.
    """
    f_fn = self.prob.potential_fn_from_dual_vec(self.g, epsilon=epsilon)
    cost_fn = self.geom.cost_fn
    return potentials.DualPotentials(f=f_fn, g=None, cost_fn=cost_fn)

  def to_dataloader(
      self, rng: jax.Array, batch_size: int, **kwargs: Any
  ) -> "semidiscrete_dataloader.SemidiscreteDataloader":
    """Create a semidiscrete dataloader.

    Args:
      rng: Random number seed used for sampling from the source distribution.
      batch_size: Batch size used in the dataloader to sample from source.
      kwargs: Keyword arguments for
        :class:`~ott.neural.data.semidiscrete_dataloader.SemidiscreteDataloader`.

    Returns:
      The semidiscrete dataloader.
    """  # noqa: E501
    from ott.neural.data import semidiscrete_dataloader

    return semidiscrete_dataloader.SemidiscreteDataloader(
        rng,
        sd_out=self,
        batch_size=batch_size,
        **kwargs,
    )

  def marginal_chi2_error(
      self,
      rng: jax.Array,
      *,
      num_iters: int,
      batch_size: int,
  ) -> jax.Array:
    """Compute the marginal chi-squared error.

    Args:
      rng: Random key used for seeding.
      num_iters: Number of iterations used to estimate the error.
      batch_size: Number of points to sample from the source distribution
        at each iteration.

    Returns:
      The marginal chi-squared error.
    """
    return _marginal_chi2_error(
        rng,
        self.g,
        self.prob,
        num_iters=num_iters,
        batch_size=batch_size,
    )

  @property
  def geom(self) -> sdpc.SemidiscretePointCloud:
    """Semidiscrete geometry."""
    return self.prob.geom


def constant_epsilon_scheduler(
    step: jax.Array, target_epsilon: jax.Array
) -> jax.Array:
  """Constant epsilon scheduler.

  Args:
    step: Current step (ignored).
    target_epsilon: Epsilon at the last iteration.

  Returns:
    The target epsilon.
  """
  del step
  return target_epsilon


@jtu.register_static
@dataclasses.dataclass(frozen=True, kw_only=True)
class SemidiscreteSolver:
  """Semidiscrete optimal transport solver.

  Args:
    num_iterations: Number of iterations.
    batch_size: Number of points to sample at each iteration.
    optimizer: Optimizer.
    error_eval_every: Compute the chi-squared error every ``error_eval_every``
      iterations.
    error_batch_size: Batch size to use when computing
      the marginal chi-squared error. If :obj:`None`, use ``batch_size``.
    error_num_repeats: Number of repeats used to estimate
      the marginal chi-squared error, set to sixteen by default.
    threshold: Convergence threshold for the marginal chi-squared error.
    potential_ema: Exponential moving average of the dual potential.
    epsilon_scheduler: Epsilon scheduler along the iterations with a signature
      ``(step, target_epsilon) -> epsilon``.
      By default, :func:`constant_epsilon_scheduler` is used.
    callback: Callback with a signature ``(state) -> None`` that is called
      at every iteration.
  """
  num_iterations: int
  batch_size: int
  optimizer: optax.GradientTransformation
  error_eval_every: int = 1000
  error_batch_size: Optional[int] = None
  error_num_repeats: int = 16
  threshold: float = 1e-3
  potential_ema: float = 0.99
  epsilon_scheduler: Callable[[jax.Array, jax.Array],
                              jax.Array] = constant_epsilon_scheduler
  callback: Optional[Callable[[SemidiscreteState], None]] = None

  def __call__(
      self,
      rng: jax.Array,
      prob: sdlp.SemidiscreteLinearProblem,
      g_init: Optional[jax.Array] = None,
  ) -> SemidiscreteOutput:
    """Run the semidiscrete solver.

    Args:
      rng: Random key used for seeding.
      prob: Semidiscrete OT problem.
      g_init: Initial potential value of shape ``[m,]``. If :obj:`None`,
        use an array of 0s.

    Returns:
      The semidiscrete output.
    """

    def cond_fn(
        it: int,
        prob: sdlp.SemidiscreteLinearProblem,
        state: SemidiscreteState,
    ) -> bool:
      loss = state.losses[it - 1]
      err = jnp.abs(state.errors[it // self.error_eval_every - 1])
      not_converged = err > self.threshold
      not_diverged = jnp.isfinite(loss)
      target_eps_not_reached = ~jnp.isclose(state.epsilon, prob.epsilon)
      # cont. if not converged and not diverged or not reached target epsilon
      return jnp.logical_or(
          it == 0,
          jnp.logical_or(
              jnp.logical_and(not_converged, not_diverged),
              target_eps_not_reached
          )
      )

    def body_fn(
        it: int,
        prob: sdlp.SemidiscreteLinearProblem,
        state: SemidiscreteState,
        compute_error: bool,
    ) -> SemidiscreteState:
      return self.step(
          jr.fold_in(rng, it),
          state=state,
          prob=prob,
          compute_error=compute_error,
          # we evaluate the error using the same samples
          rng_error=rng_error,
      )

    _, m = prob.geom.shape
    dtype = prob.geom.dtype

    if g_init is None:
      g_init = jnp.zeros(m, dtype=dtype)
    else:
      assert g_init.shape == (m,), (g_init.shape, (m,))

    state = SemidiscreteState(
        it=jnp.array(0),
        epsilon=jnp.nan,
        g=g_init,
        g_ema=g_init,
        opt_state=self.optimizer.init(g_init),
        losses=jnp.full((self.num_iterations,), fill_value=jnp.inf,
                        dtype=dtype),
        grad_norms=jnp.full((self.num_iterations,),
                            fill_value=jnp.inf,
                            dtype=dtype),
        errors=jnp.full(
            math.ceil(self.num_iterations / self.error_eval_every),
            fill_value=jnp.inf,
            dtype=dtype
        ),
    )

    rng, rng_error = jr.split(rng, 2)
    state: SemidiscreteState = fixed_point_loop.fixpoint_iter(
        cond_fn,
        body_fn,
        min_iterations=0,
        max_iterations=self.num_iterations,
        inner_iterations=self.error_eval_every,
        constants=prob,
        state=state,
    )

    return self._to_output(state, prob)

  def step(
      self,
      rng: jax.Array,
      state: SemidiscreteState,
      prob: sdlp.SemidiscreteLinearProblem,
      *,
      compute_error: bool = False,
      rng_error: Optional[jax.Array] = None,
  ) -> SemidiscreteState:
    """Perform one optimization step.

    Args:
      rng: Random seed used for sampling.
      state: Semidiscrete state.
      prob: Semidiscrete linear problem.
      compute_error: Whether to compute the marginal chi-squared error.
      rng_error: Random seed when computing the chi-squared error.

    Returns:
      The updated state.
    """
    it = state.it
    rng_error = utils.default_prng_key(rng_error)

    g_old = state.g
    epsilon = self.epsilon_scheduler(it, prob.epsilon)
    lin_prob = prob.sample(rng, self.batch_size, epsilon=epsilon)

    loss, grads = jax.value_and_grad(_semidiscrete_loss)(g_old, lin_prob)
    grad_norm = jnp.linalg.norm(grads)
    losses = state.losses.at[it].set(loss)
    grad_norms = state.grad_norms.at[it].set(grad_norm)

    updates, opt_state = self.optimizer.update(
        grads, state.opt_state, g_old, value=loss
    )
    g_new = optax.apply_updates(g_old, updates)
    g_ema = optax.incremental_update(g_new, state.g_ema, self.potential_ema)

    # fmt: off
    error = jax.lax.cond(
      compute_error,
      lambda: _marginal_chi2_error(
        rng_error, g_ema, prob,
        num_iters=self.error_num_repeats,
        batch_size=self.error_batch_size or self.batch_size,
      ),
      lambda: jnp.array(jnp.inf, dtype=state.errors.dtype),
    )
    # fmt: on
    errors = state.errors.at[it // self.error_eval_every].set(error)

    state = SemidiscreteState(
        it=it + 1,
        epsilon=epsilon,
        g=g_new,
        g_ema=g_ema,
        opt_state=opt_state,
        losses=losses,
        grad_norms=grad_norms,
        errors=errors,
    )
    if self.callback is not None:
      jax.debug.callback(self.callback, state)
    return state

  def _to_output(
      self, state: SemidiscreteState, prob: sdlp.SemidiscreteLinearProblem
  ) -> SemidiscreteOutput:
    it = state.it
    leq_thr = state.errors[it // self.error_eval_every] <= self.threshold
    finite_loss = jnp.isfinite(state.losses[it])
    return SemidiscreteOutput(
        g=state.g_ema,
        prob=prob,
        it=it,
        losses=state.losses,
        errors=state.errors,
        converged=jnp.logical_and(leq_thr, finite_loss),
    )


@jax.custom_vjp
def _semidiscrete_loss(
    g: jax.Array,
    prob: linear_problem.LinearProblem,
) -> jax.Array:
  f, _ = prob._c_transform(g, axis=1)
  # we assume uniform weights for `prob.a`
  return -(jnp.mean(f) + jnp.dot(g, prob.b))


def _semidiscrete_loss_fwd(
    g: jax.Array,
    prob: linear_problem.LinearProblem,
) -> Tuple[jax.Array, Tuple[jax.Array, linear_problem.LinearProblem]]:
  f, z = prob._c_transform(g, axis=1)
  # we assume uniform weights for `prob.a`
  return -(jnp.mean(f) + jnp.dot(g, prob.b)), (z, prob)


def _semidiscrete_loss_bwd(
    res: jax.Array,
    g: jax.Array,
) -> Tuple[jax.Array, None]:

  def soft_grad(z: jax.Array) -> jax.Array:
    if prob._b is None:  # uniform weights
      return jsp.special.softmax(z, axis=-1).sum(0)
    return _weighted_softmax(z, b=prob.b, axis=-1).sum(0)

  def hard_grad(z: jax.Array) -> jax.Array:
    pos_weights = prob.b[None, :] > 0.0
    ixs = jnp.argmax(jnp.where(pos_weights, z, -jnp.inf), axis=-1)
    return jax.ops.segment_sum(
        jnp.ones(n, dtype=z.dtype), segment_ids=ixs, num_segments=m
    )

  z, prob = res
  is_soft = prob.geom.epsilon > 0.0
  n, m = prob.geom.shape

  grad = jax.lax.cond(is_soft, soft_grad, hard_grad, z)
  assert grad.shape == (m,), (grad.shape, (m,))

  grad = grad * (1.0 / n) - prob.b
  return g * grad, None


def _weighted_softmax(x: jax.Array, b: jax.Array, axis: int = -1) -> jax.Array:
  where = b > 0.0
  x_max = jnp.max(x, axis=axis, keepdims=True, where=where, initial=-jnp.inf)
  unnormalized = b * jnp.exp(x - x_max)
  softmax = unnormalized / unnormalized.sum(
      axis=axis, keepdims=True, where=where
  )
  return jnp.where(where, softmax, 0.0)


_semidiscrete_loss.defvjp(_semidiscrete_loss_fwd, _semidiscrete_loss_bwd)


def _marginal_chi2_error(
    rng: jax.Array,
    g: jax.Array,
    prob: sdlp.SemidiscreteLinearProblem,
    *,
    num_iters: int,
    batch_size: int,
) -> jax.Array:

  def compute_chi2(matrix: Union[jax.Array, jesp.BCOO]) -> jax.Array:
    """Compute chi2 metric.

    Implements Eq. 3.5  in https://arxiv.org/pdf/2509.25519v1,
    from reference :cite:`mousavi:25`.

    Args:
      matrix: coupling matrix, either dense or sparse.

    Returns:
      Chi2 metric estimator, lowerbounded by -1.0
    """
    # normalize coupling matrix to have columns sum to 1.
    matrix = batch_size * matrix

    if isinstance(matrix, jesp.BCOO):
      out = jesp.bcoo_reduce_sum(matrix, axes=(0,)).todense() ** 2
      out -= jesp.bcoo_reduce_sum(matrix ** 2, axes=(0,)).todense()
    else:
      out = jnp.sum(matrix, axis=0) ** 2 - jnp.sum(matrix ** 2, axis=0)

    out = jnp.sum(out / prob.b) / (batch_size * (batch_size - 1.0))
    return out - 1.0

  def body(chi2_err_avg: jax.Array, it: jax.Array) -> Tuple[jax.Array, None]:
    rng_it = jr.fold_in(rng, it)
    matrix = out.sample(rng_it, batch_size).matrix
    chi2 = compute_chi2(matrix)
    chi2_err_avg = chi2_err_avg + chi2 / num_iters
    return chi2_err_avg, None

  out = SemidiscreteOutput(g=g, prob=prob)

  chi2_err = jnp.zeros((), dtype=g.dtype)
  chi2_err, _ = jax.lax.scan(body, init=chi2_err, xs=jnp.arange(num_iters))
  return chi2_err



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Tuple

import jax.numpy as jnp

from optax import assignment

from ott.geometry import costs, geometry, pointcloud
from ott.problems.linear import linear_problem
from ott.solvers.linear import semidiscrete

__all__ = ["hungarian", "wassdis_p"]


def hungarian(
    geom: geometry.Geometry
) -> Tuple[jnp.ndarray, semidiscrete.HardAssignmentOutput]:
  """Solve matching problem using the :term:`Hungarian algorithm`.

  Uses the implementation from :mod:`optax`.

  Args:
    geom: Geometry object with square (shape ``[n, n]``)
      :attr:`~ott.geometry.geometry.Geometry.cost_matrix`.

  Returns:
    The value of the unregularized OT problem, along with an output
    object listing relevant information on outputs.
  """
  n, m = geom.shape
  assert n == m, f"Hungarian can only match same # of points, got {n} and {m}."
  cost_matrix = geom.cost_matrix
  i, j = assignment.hungarian_algorithm(cost_matrix)
  prob = linear_problem.LinearProblem(geom)
  out = semidiscrete.HardAssignmentOutput(
      prob, paired_indices=jnp.stack([i, j])
  )
  transport_cost = cost_matrix[i, j].sum() / n
  return transport_cost, out


def wassdis_p(x: jnp.ndarray, y: jnp.ndarray, *, p: float = 2.0) -> float:
  """Compute the :term:`Wasserstein distance`, uses :term:`Hungarian algorithm`.

  Uses :func:`hungarian` to solve the :term:`optimal matching problem` between
  two point clouds of the same size, to compute a :term:`Wasserstein distance`
  estimator.

  Note:
    At the moment, only supports point clouds of the same size to be easily
    cast as an optimal matching problem.

  Args:
    x: ``[n, d]`` point cloud.
    y: ``[n, d]`` point cloud of the same size.
    p: order of the Wasserstein distance, non-negative float.

  Returns:
    The `p`-Wasserstein distance between these point clouds.
  """
  geom = pointcloud.PointCloud(x, y, cost_fn=costs.EuclideanP(p=p))
  cost, _ = hungarian(geom)
  return cost ** (1.0 / p)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import jax
import jax.numpy as jnp

from flax import nnx

__all__ = ["EMA", "init_ema", "update_ema"]


class EMA(nnx.Module):
  """Exponential moving average (EMA) of a model.

  Args:
    model: Model to average.
    decay: EMA decay factor.
  """

  def __init__(self, model: nnx.Module, *, decay: float):
    super().__init__()
    self.ema = init_ema(model)
    self.decay = decay

  def __call__(self, model: nnx.Module) -> None:
    """Update the EMA.

    Args:
      model: Model to average.

    Returns:
      Nothing, just updates the EMA in-place.
    """
    update_ema(model, ema=self.ema, decay=self.decay)


def init_ema(model: nnx.Module) -> nnx.Module:
  """Create initial exponential moving average (EMA) state.

  Args:
    model: Model to average.

  Returns:
    Copy of the model with parameters set to 0s.
  """
  graphdef, state, rest = nnx.split(model, nnx.Param, ...)
  ema_state = jax.tree.map(jnp.zeros_like, state)
  # copy rest of the params, like RNGs, batch stats, etc.
  rest = jax.tree.map(lambda r: r.copy(), rest)
  return nnx.merge(graphdef, ema_state, rest)


def update_ema(model: nnx.Module, *, ema: nnx.Module, decay: float) -> None:
  """Update the EMA of a model.

  Args:
    model: Model to average.
    ema: EMA of the model.
    decay: Decay factor.

  Returns:
    Nothing, just updates the EMA in-place.
  """

  def update_fn(p_model: nnx.Param, p_ema: nnx.Param) -> nnx.Param:
    return p_ema * decay + p_model * (1.0 - decay)

  state, rest = nnx.state(model, nnx.Param, ...)
  graphdef, ema_state, _ = nnx.split(ema, nnx.Param, ...)
  rest = jax.tree.map(lambda r: r.copy(), rest)
  ema_state = jax.tree.map(update_fn, state, ema_state)
  nnx.update(ema, ema_state, rest)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Callable, Optional

import jax
import jax.numpy as jnp
import jax.tree_util as jtu

from ott.geometry import semidiscrete_pointcloud
from ott.problems.linear import linear_problem

__all__ = ["SemidiscreteLinearProblem"]


@jtu.register_pytree_node_class
class SemidiscreteLinearProblem:
  """Semidiscrete linear OT problem.

  Instances of this problem can be sampled using the :meth:`sample` method.

  Args:
    geom: Semidiscrete point cloud geometry.
    b: The second marginal. If :obj:`None`, it will be uniform.
    tau_b: If :math:`< 1`, defines how much unbalanced the problem is
      on the second marginal. Currently not implemented.
  """

  def __init__(
      self,
      geom: semidiscrete_pointcloud.SemidiscretePointCloud,
      b: Optional[jax.Array] = None,
      tau_b: float = 1.0,
  ):
    assert tau_b == 1.0, "Unbalanced semidiscrete problem is not supported."
    self.geom = geom
    self._b = b
    self.tau_b = tau_b

  def sample(
      self,
      rng: jax.Array,
      num_samples: int,
      *,
      epsilon: Optional[float] = None,
  ) -> linear_problem.LinearProblem:
    """Sample a linear OT problem.

    Args:
      rng: Random key used for seeding.
      num_samples: Number of samples.
      epsilon: Epsilon regularization. If :obj:`None`, use :attr:`epsilon`.

    Returns:
      The sampled linear problem.
    """
    if epsilon is None:
      epsilon = self.epsilon
    geom = self.geom.sample(rng, num_samples, epsilon=epsilon)
    return linear_problem.LinearProblem(
        geom, a=None, b=self._b, tau_a=1.0, tau_b=self.tau_b
    )

  def potential_fn_from_dual_vec(
      self,
      g: jax.Array,
      *,
      epsilon: Optional[float] = None
  ) -> Callable[[jax.Array], jax.Array]:
    r"""Get potential function from a dual vector using the :term:`c-transform`.

    Args:
      g: Potential vector :math:`\mathbb{g}` of shape ``[m,]``.
      epsilon: Epsilon regularization. If :obj:`None`, use in the :attr:`geom`.

    Returns:
      The dual potential function :math:`f`.
    """
    # `potential_fn_from_dual_vec` accesses only necessary properties of the
    # problem/geometry, so we can pass the semidiscrete point cloud
    prob = linear_problem.LinearProblem(self.geom, b=self.b)
    return prob.potential_fn_from_dual_vec(g, epsilon=epsilon, axis=1)

  @property
  def b(self) -> jnp.ndarray:
    """Second marginal."""
    if self._b is not None:
      return self._b
    _, m = self.geom.shape
    return jnp.full((m,), fill_value=1.0 / m, dtype=self.geom.y.dtype)

  @property
  def epsilon(self) -> jax.Array:
    """Entropic regularization."""
    return self.geom.epsilon

  def tree_flatten(self):  # noqa: D102
    return (self.geom, self._b), {"tau_b": self.tau_b}

  @classmethod
  def tree_unflatten(  # noqa: D102
      cls, aux_data, children
  ) -> "SemidiscreteLinearProblem":
    return cls(*children, **aux_data)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A Jax implementation of the unbalanced low-rank GW algorithm."""
from typing import Any, Callable, Mapping, NamedTuple, Optional, Tuple

import jax
import jax.numpy as jnp
import jax.scipy as jsp
import numpy as np

from ott import utils
from ott.geometry import geometry, low_rank
from ott.initializers.linear import initializers_lr
from ott.math import fixed_point_loop
from ott.math import unbalanced_functions as uf
from ott.math import utils as mu
from ott.problems.quadratic import quadratic_problem
from ott.solvers.linear import lr_utils, sinkhorn

__all__ = ["LRGromovWasserstein", "LRGWOutput"]

ProgressFunction = Callable[
    [Tuple[np.ndarray, np.ndarray, np.ndarray, "LRGWState"]], None]


class LRGWState(NamedTuple):
  """State of the low-rank GW algorithm."""
  q: jnp.ndarray
  r: jnp.ndarray
  g: jnp.ndarray
  gamma: float
  costs: jnp.ndarray
  errors: jnp.ndarray
  crossed_threshold: bool

  def compute_error(  # noqa: D102
      self, previous_state: "LRGWState"
  ) -> float:
    err_q = mu.gen_js(self.q, previous_state.q, c=1.0)
    err_r = mu.gen_js(self.r, previous_state.r, c=1.0)
    err_g = mu.gen_js(self.g, previous_state.g, c=1.0)

    return ((1.0 / self.gamma) ** 2) * (err_q + err_r + err_g)

  def reg_gw_cost(  # noqa: D102
    self,
    ot_prob: quadratic_problem.QuadraticProblem,
    *,
    epsilon: float,
    use_danskin: bool = False
  ) -> float:
    return compute_reg_gw_cost(
        self.q,
        self.r,
        self.g,
        ot_prob,
        epsilon=epsilon,
        use_danskin=use_danskin
    )

  def set(self, **kwargs: Any) -> "LRGWState":
    """Return a copy of self, with potential overwrites."""
    return self._replace(**kwargs)


def compute_reg_gw_cost(
    q: jnp.ndarray,
    r: jnp.ndarray,
    g: jnp.ndarray,
    ot_prob: quadratic_problem.QuadraticProblem,
    epsilon: float,
    use_danskin: bool = False
) -> float:
  """Compute the regularized OT cost, here the primal cost of the LR solution.

  Args:
    q: first factor of solution
    r: second factor of solution
    g: weights of solution
    ot_prob: linear problem
    epsilon: Entropic regularization.
    use_danskin: if True, use Danskin's theorem :cite:`danskin:67,bertsekas:71`
      to avoid computing the gradient of the cost function.

  Returns:
    regularized OT cost, the (primal) transport cost of the low-rank solution.
  """

  def ent(x: jnp.ndarray) -> float:
    # generalized entropy
    return jnp.sum(jsp.special.entr(x) + x)

  q = jax.lax.stop_gradient(q) if use_danskin else q
  r = jax.lax.stop_gradient(r) if use_danskin else r
  g = jax.lax.stop_gradient(g) if use_danskin else g

  out = LRGWOutput(
      q=q,
      r=r,
      g=g,
      ot_prob=ot_prob,
      costs=None,
      errors=None,
      epsilon=None,
      inner_iterations=None,
      converged=False,
  )

  cost = out.primal_cost - epsilon * (ent(q) + ent(r) + ent(g))
  if ot_prob.tau_a != 1.0:
    rho_a = uf.rho(1.0, ot_prob.tau_a)
    cost += rho_a * mu.gen_kl(jnp.sum(q, axis=1), ot_prob.a)
  if ot_prob.tau_b != 1.0:
    rho_b = uf.rho(1.0, ot_prob.tau_b)
    cost += rho_b * mu.gen_kl(jnp.sum(r, axis=1), ot_prob.b)

  return cost


class LRGWOutput(NamedTuple):
  """Transport interface for a low-rank GW solution."""
  q: jnp.ndarray
  r: jnp.ndarray
  g: jnp.ndarray
  costs: jnp.ndarray
  # TODO(michalk8): must be called `errors`, because of `store_inner_errors`
  # in future, enforce via class hierarchy
  errors: jnp.ndarray
  ot_prob: quadratic_problem.QuadraticProblem
  epsilon: float
  inner_iterations: int
  converged: bool
  reg_gw_cost: Optional[float] = None

  def set(self, **kwargs: Any) -> "LRGWOutput":
    """Return a copy of self, with potential overwrites."""
    return self._replace(**kwargs)

  def set_cost(  # noqa: D102
    self,
    ot_prob: quadratic_problem.QuadraticProblem,
    lse_mode: bool,
    use_danskin: bool = False
  ) -> "LRGWOutput":
    del lse_mode
    return self.set(reg_gw_cost=self.compute_reg_gw_cost(ot_prob, use_danskin))

  def compute_reg_gw_cost(  # noqa: D102
    self,
    ot_prob: quadratic_problem.QuadraticProblem,
    use_danskin: bool = False,
  ) -> float:
    return compute_reg_gw_cost(
        self.q,
        self.r,
        self.g,
        ot_prob,
        epsilon=self.epsilon,
        use_danskin=use_danskin
    )

  @property
  def geom(self) -> geometry.Geometry:  # noqa: D102
    """Linearized geometry."""
    return _linearized_geometry(self.ot_prob, q=self.q, r=self.r, g=self.g)

  @property
  def a(self) -> jnp.ndarray:  # noqa: D102
    return self.ot_prob.a

  @property
  def b(self) -> jnp.ndarray:  # noqa: D102
    return self.ot_prob.b

  @property
  def n_iters(self) -> int:  # noqa: D102
    return jnp.sum(self.errors != -1) * self.inner_iterations

  @property
  def matrix(self) -> jnp.ndarray:
    """Transport matrix if it can be instantiated."""
    return (self.q * self._inv_g) @ self.r.T

  def apply(self, inputs: jnp.ndarray, axis: int = 0) -> jnp.ndarray:
    """Apply the transport to a array; axis=1 for its transpose."""
    q, r = (self.q, self.r) if axis == 1 else (self.r, self.q)
    # for `axis=0`: (batch, m), (m, r), (r,), (r, n)
    return ((inputs @ r) * self._inv_g) @ q.T

  def marginal(self, axis: int) -> jnp.ndarray:  # noqa: D102
    length = self.q.shape[0] if axis == 0 else self.r.shape[0]
    return self.apply(jnp.ones(length,), axis=axis)

  def cost_at_geom(self, other_geom: geometry.Geometry) -> float:
    """Return OT cost for current solution, evaluated at any cost matrix."""
    return jnp.sum(self.q * other_geom.apply_cost(self.r, axis=1) * self._inv_g)

  def transport_cost_at_geom(self, other_geom: geometry.Geometry) -> float:
    """Return (by recomputing it) bare transport cost of current solution."""
    return self.cost_at_geom(other_geom)

  @property
  def primal_cost(self) -> float:
    """Return (by recomputing it) transport cost of current solution."""
    geom_xx, geom_yy = self.ot_prob.geom_xx, self.ot_prob.geom_yy
    marginal_a = self.ot_prob.a if self.ot_prob.tau_a == 1.0 else self.q.sum(1)
    marginal_b = self.ot_prob.b if self.ot_prob.tau_b == 1.0 else self.r.sum(1)

    quad_cost = 0.5 * self.transport_cost_at_geom(other_geom=self.geom)
    quad_cost += jnp.vdot(geom_xx.apply_square_cost(marginal_a), marginal_a)
    quad_cost += jnp.vdot(geom_yy.apply_square_cost(marginal_b), marginal_b)

    if not self.ot_prob.is_fused:
      return quad_cost

    alpha = self.ot_prob.fused_penalty / (self.ot_prob.fused_penalty + 1.0)
    norm_g = jnp.linalg.norm(self.g, ord=1)

    lin_cost = self.cost_at_geom(self.ot_prob.geom_xy)
    return alpha * norm_g * lin_cost + (1.0 - alpha) * quad_cost

  @property
  def transport_mass(self) -> float:
    """Sum of transport matrix."""
    return self.marginal(0).sum()

  @property
  def _inv_g(self) -> jnp.ndarray:
    return 1.0 / self.g


@jax.tree_util.register_pytree_node_class
class LRGromovWasserstein(sinkhorn.Sinkhorn):
  r"""Low-rank Gromov-Wasserstein solver :cite:`scetbon:23`.

  The algorithm minimizes a non-convex problem. It therefore requires special
  care to initialization and convergence. Convergence is evaluated on successive
  evaluations of the objective.

  .. warning::
    This solver only for the **unbalanced** case. Balanced case is implemented
    in :class:`~ott.solvers.quadratic.gromov_wasserstein.GromovWasserstein`
    and will be unified here in the future release.

  Args:
    rank: Rank constraint on the coupling to minimize the linear OT problem
    gamma: The (inverse of) gradient step size used by mirror descent.
    gamma_rescale: Whether to rescale :math:`\gamma` every iteration as
      described in :cite:`scetbon:22b`.
    epsilon: Entropic regularization added on top of low-rank problem.
    initializer: How to initialize the :math:`Q`, :math:`R` and :math:`g`
      factors. If :obj:`None`, use
      :class:`~ott.initializers.linear.initializers_lr.RandomInitializer`.
    lse_mode: Whether to run computations in LSE or kernel mode.
    inner_iterations: Number of inner iterations used by the algorithm before
      re-evaluating progress.
    min_iterations: The minimum number of low-rank Sinkhorn iterations carried
      out before the error is computed and monitored.
    max_iterations: The maximum number of low-rank Sinkhorn iterations.
    use_danskin: Use Danskin theorem to evaluate gradient of objective w.r.t.
      input parameters. Only `True` handled at this moment.
    implicit_diff: Whether to use implicit differentiation. Currently, only
      ``implicit_diff = False`` is implemented.
    progress_fn: callback function which gets called during the GW
      iterations, so the user can display the error at each iteration,
      e.g., using a progress bar. See :func:`~ott.utils.default_progress_fn`
      for a basic implementation.
    kwargs_dys: Keyword arguments passed to :meth:`dykstra_update_lse`,
      :meth:`dykstra_update_kernel` or one of the functions defined in
      :mod:`ott.solvers.linear`, depending on the ``lse_mode``.
    kwargs: Keyword arguments for
      :class:`~ott.solvers.linear.sinkhorn.Sinkhorn`.
  """

  def __init__(
      self,
      rank: int,
      gamma: float = 10.0,
      gamma_rescale: bool = True,
      epsilon: float = 0.0,
      initializer: Optional[initializers_lr.LRInitializer] = None,
      lse_mode: bool = True,
      use_danskin: bool = True,
      implicit_diff: bool = False,
      inner_iterations: int = 2_000,
      min_iterations: int = 10_000,
      max_iterations: int = 100_000,
      kwargs_dys: Optional[Mapping[str, Any]] = None,
      progress_fn: Optional[ProgressFunction] = None,
      **kwargs: Any,
  ):
    assert not implicit_diff, "Implicit diff. not yet implemented."
    super().__init__(
        lse_mode=lse_mode,
        inner_iterations=inner_iterations,
        min_iterations=min_iterations,
        max_iterations=max_iterations,
        use_danskin=use_danskin,
        implicit_diff=implicit_diff,
        **kwargs
    )
    self.rank = rank
    self.gamma = gamma
    self.gamma_rescale = gamma_rescale
    self.epsilon = epsilon
    self.initializer = initializers_lr.RandomInitializer(
        rank
    ) if initializer is None else initializer
    self.progress_fn = progress_fn
    # can be `None`
    self.kwargs_dys = {} if kwargs_dys is None else kwargs_dys

  def __call__(
      self,
      ot_prob: quadratic_problem.QuadraticProblem,
      init: Optional[Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]] = None,
      rng: Optional[jax.Array] = None,
      **kwargs: Any,
  ) -> LRGWOutput:
    """Run the low-rank Gromov-Wasserstein solver.

    Args:
      ot_prob: Quadratic OT problem.
      init: Initial values of the low-rank factors:

        - :attr:`~ott.solvers.linear.sinkhorn_lr.LRGWOutput.q`.
        - :attr:`~ott.solvers.linear.sinkhorn_lr.LRGWOutput.r`.
        - :attr:`~ott.solvers.linear.sinkhorn_lr.LRGWOutput.g`.

        If :obj:`None`, use the initializer.
      rng: Random key for seeding.
      kwargs: Keyword arguments for the initializer.

    Returns:
      The low-rank GW output.
    """
    rng = utils.default_prng_key(rng)
    rng_lrc, rng_init = jax.random.split(rng)

    if ot_prob._is_low_rank_convertible:
      ot_prob = ot_prob.to_low_rank(rng=rng_lrc)
    if init is None:
      init = self.initializer(ot_prob, rng=rng_init, **kwargs)

    return run(ot_prob, self, init)

  def _get_costs(
      self,
      ot_prob: quadratic_problem.QuadraticProblem,
      state: LRGWState,
  ) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, float]:
    q, r, g = state.q, state.r, state.g
    log_q, log_r, log_g = mu.safe_log(q), mu.safe_log(r), mu.safe_log(g)
    inv_g = 1.0 / g[None, :]
    lin_geom = _linearized_geometry(ot_prob, q=q, r=r, g=g)

    tmp = lin_geom.apply_cost(r, axis=1)
    grad_q = tmp * inv_g
    if ot_prob.tau_a != 1.0:  # unbalanced grad
      grad_q += 2.0 * ot_prob.geom_xx.apply_square_cost(
          q.sum(1), axis=1
      )[:, None]

    grad_r = lin_geom.apply_cost(q, axis=0) * inv_g
    if ot_prob.tau_b != 1.0:  # unbalanced grad
      grad_r += 2.0 * ot_prob.geom_yy.apply_square_cost(
          r.sum(1), axis=1
      )[:, None]

    omega_quad = jnp.sum(q * tmp, axis=0)
    grad_g = -omega_quad / (g ** 2)

    if ot_prob.is_fused:
      alpha = ot_prob.fused_penalty / (ot_prob.fused_penalty + 1.0)
      norm_g = jnp.linalg.norm(g, ord=1)

      tmp = ot_prob.geom_xy.apply_cost(r, axis=1)
      lin_grad_q = tmp * inv_g * norm_g
      lin_grad_r = ot_prob.geom_xy.apply_cost(q) * inv_g * norm_g

      omega_lin = jnp.sum(q * tmp, axis=0)
      lin_grad_g = -omega_lin / (g ** 2) * norm_g + jnp.sum(q * tmp * inv_g)

      grad_q = alpha * lin_grad_q + (1.0 - alpha) * grad_q
      grad_r = alpha * lin_grad_r + (1.0 - alpha) * grad_r
      grad_g = alpha * lin_grad_g + (1.0 - alpha) * grad_g

    grad_q += self.epsilon * log_q
    grad_r += self.epsilon * log_r
    grad_g += self.epsilon * log_g

    if self.gamma_rescale:
      norm_q = jnp.max(jnp.abs(grad_q)) ** 2
      norm_r = jnp.max(jnp.abs(grad_r)) ** 2
      norm_g = jnp.max(jnp.abs(grad_g)) ** 2
      gamma = self.gamma / jnp.max(jnp.array([norm_q, norm_r, norm_g]))
    else:
      gamma = self.gamma

    eps_factor = 1.0 / (self.epsilon * gamma + 1.0)
    gamma *= eps_factor

    c_q = -gamma * grad_q + eps_factor * log_q
    c_r = -gamma * grad_r + eps_factor * log_r
    c_g = -gamma * grad_g + eps_factor * log_g

    return c_q, c_r, c_g, gamma

  # TODO(michalk8): move to `lr_utils` when refactoring this the future
  def dykstra_update_lse(
      self,
      c_q: jnp.ndarray,
      c_r: jnp.ndarray,
      h: jnp.ndarray,
      gamma: float,
      ot_prob: quadratic_problem.QuadraticProblem,
      min_entry_value: float = 1e-6,
      tolerance: float = 1e-3,
      min_iter: int = 0,
      inner_iter: int = 10,
      max_iter: int = 10000
  ) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Run Dykstra's algorithm."""
    # shortcuts for problem's definition.
    r = self.rank
    n, m = ot_prob.geom_xx.shape[0], ot_prob.geom_yy.shape[0]
    loga, logb = jnp.log(ot_prob.a), jnp.log(ot_prob.b)

    h_old = h
    g1_old, g2_old = jnp.zeros(r), jnp.zeros(r)
    f1, f2 = jnp.zeros(n), jnp.zeros(m)

    w_gi, w_gp = jnp.zeros(r), jnp.zeros(r)
    w_q, w_r = jnp.zeros(r), jnp.zeros(r)
    err = jnp.inf
    state_inner = f1, f2, g1_old, g2_old, h_old, w_gi, w_gp, w_q, w_r, err
    constants = c_q, c_r, loga, logb

    def cond_fn(
        iteration: int, constants: Tuple[jnp.ndarray, ...],
        state_inner: Tuple[jnp.ndarray, ...]
    ) -> bool:
      del iteration, constants
      *_, err = state_inner
      return err > tolerance

    def _softm(
        f: jnp.ndarray, g: jnp.ndarray, c: jnp.ndarray, axis: int
    ) -> jnp.ndarray:
      return jsp.special.logsumexp(
          gamma * (f[:, None] + g[None, :] - c), axis=axis
      )

    def body_fn(
        iteration: int, constants: Tuple[jnp.ndarray, ...],
        state_inner: Tuple[jnp.ndarray, ...], compute_error: bool
    ) -> Tuple[jnp.ndarray, ...]:
      # TODO(michalk8): in the future, use `NamedTuple`
      f1, f2, g1_old, g2_old, h_old, w_gi, w_gp, w_q, w_r, err = state_inner
      c_q, c_r, loga, logb = constants

      # First Projection
      f1 = jnp.where(
          jnp.isfinite(loga),
          (loga - _softm(f1, g1_old, c_q, axis=1)) / gamma + f1, loga
      )
      f2 = jnp.where(
          jnp.isfinite(logb),
          (logb - _softm(f2, g2_old, c_r, axis=1)) / gamma + f2, logb
      )

      h = h_old + w_gi
      h = jnp.maximum(jnp.log(min_entry_value) / gamma, h)
      w_gi += h_old - h
      h_old = h

      # Update couplings
      g_q = _softm(f1, g1_old, c_q, axis=0)
      g_r = _softm(f2, g2_old, c_r, axis=0)

      # Second Projection
      h = (1.0 / 3.0) * (h_old + w_gp + w_q + w_r)
      h += g_q / (3.0 * gamma)
      h += g_r / (3.0 * gamma)
      g1 = h + g1_old - g_q / gamma
      g2 = h + g2_old - g_r / gamma

      w_q = w_q + g1_old - g1
      w_r = w_r + g2_old - g2
      w_gp = h_old + w_gp - h

      q, r, _ = recompute_couplings(f1, g1, c_q, f2, g2, c_r, h, gamma)

      g1_old = g1
      g2_old = g2
      h_old = h

      err = jax.lax.cond(
          jnp.logical_and(compute_error, iteration >= min_iter),
          lambda: dykstra_solution_error(q, r, ot_prob, self.norm_error)[0],
          lambda: err
      )

      return f1, f2, g1_old, g2_old, h_old, w_gi, w_gp, w_q, w_r, err

    def recompute_couplings(
        f1: jnp.ndarray,
        g1: jnp.ndarray,
        c_q: jnp.ndarray,
        f2: jnp.ndarray,
        g2: jnp.ndarray,
        c_r: jnp.ndarray,
        h: jnp.ndarray,
        gamma: float,
    ) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
      q = jnp.exp(gamma * (f1[:, None] + g1[None, :] - c_q))
      r = jnp.exp(gamma * (f2[:, None] + g2[None, :] - c_r))
      g = jnp.exp(gamma * h)
      return q, r, g

    state_inner = fixed_point_loop.fixpoint_iter_backprop(
        cond_fn, body_fn, min_iter, max_iter, inner_iter, constants, state_inner
    )

    f1, f2, g1_old, g2_old, h_old, _, _, _, _, _ = state_inner
    return recompute_couplings(f1, g1_old, c_q, f2, g2_old, c_r, h_old, gamma)

  def dykstra_update_kernel(
      self,
      k_q: jnp.ndarray,
      k_r: jnp.ndarray,
      k_g: jnp.ndarray,
      gamma: float,
      ot_prob: quadratic_problem.QuadraticProblem,
      min_entry_value: float = 1e-6,
      tolerance: float = 1e-3,
      min_iter: int = 0,
      inner_iter: int = 10,
      max_iter: int = 10000
  ) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Run Dykstra's algorithm."""
    # shortcuts for problem's definition.
    del gamma
    rank = self.rank
    n, m = ot_prob.geom_xx.shape[0], ot_prob.geom_yy.shape[0]
    a, b = ot_prob.a, ot_prob.b
    supp_a, supp_b = a > 0, b > 0

    g_old = k_g
    v1_old, v2_old = jnp.ones(rank), jnp.ones(rank)
    u1, u2 = jnp.ones(n), jnp.ones(m)

    q_gi, q_gp = jnp.ones(rank), jnp.ones(rank)
    q_q, q_r = jnp.ones(rank), jnp.ones(rank)
    err = jnp.inf
    state_inner = u1, u2, v1_old, v2_old, g_old, q_gi, q_gp, q_q, q_r, err
    constants = k_q, k_r, k_g, a, b

    def cond_fn(
        iteration: int, constants: Tuple[jnp.ndarray, ...],
        state_inner: Tuple[jnp.ndarray, ...]
    ) -> bool:
      del iteration, constants
      *_, err = state_inner
      return err > tolerance

    def body_fn(
        iteration: int, constants: Tuple[jnp.ndarray, ...],
        state_inner: Tuple[jnp.ndarray, ...], compute_error: bool
    ) -> Tuple[jnp.ndarray, ...]:
      # TODO(michalk8): in the future, use `NamedTuple`
      u1, u2, v1_old, v2_old, g_old, q_gi, q_gp, q_q, q_r, err = state_inner
      k_q, k_r, k_g, a, b = constants

      # First Projection
      u1 = jnp.where(supp_a, a / jnp.dot(k_q, v1_old), 0.0)
      u2 = jnp.where(supp_b, b / jnp.dot(k_r, v2_old), 0.0)
      g = jnp.maximum(min_entry_value, g_old * q_gi)
      q_gi = (g_old * q_gi) / g
      g_old = g

      # Second Projection
      v1_trans = jnp.dot(k_q.T, u1)
      v2_trans = jnp.dot(k_r.T, u2)
      g = (g_old * q_gp * v1_old * q_q * v1_trans * v2_old * q_r *
           v2_trans) ** (1 / 3)
      v1 = g / v1_trans
      v2 = g / v2_trans
      q_gp = (g_old * q_gp) / g
      q_q = (q_q * v1_old) / v1
      q_r = (q_r * v2_old) / v2
      v1_old = v1
      v2_old = v2
      g_old = g

      # Compute Couplings
      q, r, _ = recompute_couplings(u1, v1, k_q, u2, v2, k_r, g)

      err = jax.lax.cond(
          jnp.logical_and(compute_error, iteration >= min_iter),
          lambda: dykstra_solution_error(q, r, ot_prob, self.norm_error)[0],
          lambda: err
      )

      return u1, u2, v1_old, v2_old, g_old, q_gi, q_gp, q_q, q_r, err

    def recompute_couplings(
        u1: jnp.ndarray,
        v1: jnp.ndarray,
        k_q: jnp.ndarray,
        u2: jnp.ndarray,
        v2: jnp.ndarray,
        k_r: jnp.ndarray,
        g: jnp.ndarray,
    ) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
      q = u1.reshape((-1, 1)) * k_q * v1.reshape((1, -1))
      r = u2.reshape((-1, 1)) * k_r * v2.reshape((1, -1))
      return q, r, g

    state_inner = fixed_point_loop.fixpoint_iter_backprop(
        cond_fn, body_fn, min_iter, max_iter, inner_iter, constants, state_inner
    )

    u1, u2, v1_old, v2_old, g_old, _, _, _, _, _ = state_inner
    return recompute_couplings(u1, v1_old, k_q, u2, v2_old, k_r, g_old)

  def lse_step(
      self, ot_prob: quadratic_problem.QuadraticProblem, state: LRGWState,
      iteration: int
  ) -> LRGWState:
    """Low-rank GW LSE update."""
    c_q, c_r, c_g, gamma = self._get_costs(ot_prob, state)

    if ot_prob.is_balanced:
      c_q, c_r, h = c_q / -gamma, c_r / -gamma, c_g / gamma
      q, r, g = self.dykstra_update_lse(
          c_q, c_r, h, gamma, ot_prob, **self.kwargs_dys
      )
    else:
      q, r, g = lr_utils.unbalanced_dykstra_lse(
          c_q, c_r, c_g, gamma, ot_prob, **self.kwargs_dys
      )
    return state.set(q=q, g=g, r=r, gamma=gamma)  #, (c_q, c_r, c_g)

  def kernel_step(
      self, ot_prob: quadratic_problem.QuadraticProblem, state: LRGWState,
      iteration: int
  ) -> LRGWState:
    """Low-rank GW kernel update."""
    c_q, c_r, c_g, gamma = self._get_costs(ot_prob, state)
    c_q, c_r, c_g = jnp.exp(c_q), jnp.exp(c_r), jnp.exp(c_g)

    if ot_prob.is_balanced:
      q, r, g = self.dykstra_update_kernel(
          c_q, c_r, c_g, gamma, ot_prob, **self.kwargs_dys
      )
    else:
      q, r, g = lr_utils.unbalanced_dykstra_kernel(
          c_q, c_r, c_g, gamma, ot_prob, **self.kwargs_dys
      )
    return state.set(q=q, g=g, r=r, gamma=gamma)

  def one_iteration(
      self, ot_prob: quadratic_problem.QuadraticProblem, state: LRGWState,
      iteration: int, compute_error: bool
  ) -> LRGWState:
    """Carries out one low-rank GW iteration.

    Depending on lse_mode, these iterations can be either in:

      - log-space for numerical stability.
      - scaling space, using standard kernel-vector multiply operations.

    Args:
      ot_prob: the transport problem definition
      state: the current state.
      iteration: the current iteration of the GW outer loop.
      compute_error: flag to indicate this iteration computes/stores an error

    Returns:
      The updated state.
    """
    previous_state = state

    it = iteration // self.inner_iterations
    if self.lse_mode:  # In lse_mode, run additive updates.
      state = self.lse_step(ot_prob, state, iteration)
    else:
      state = self.kernel_step(ot_prob, state, iteration)

    # re-computes error if compute_error is True, else set it to inf.
    cost = jax.lax.cond(
        jnp.logical_and(compute_error, iteration >= self.min_iterations),
        lambda: state.reg_gw_cost(ot_prob, epsilon=self.epsilon),
        lambda: jnp.inf
    )
    error = jax.lax.cond(
        iteration >= self.min_iterations,
        lambda: state.compute_error(previous_state), lambda: jnp.inf
    )
    crossed_threshold = jnp.logical_or(
        state.crossed_threshold,
        jnp.logical_and(
            state.errors[it - 1] >= self.threshold, error < self.threshold
        )
    )

    state = state.set(
        costs=state.costs.at[it].set(cost),
        errors=state.errors.at[it].set(error),
        crossed_threshold=crossed_threshold,
    )

    if self.progress_fn is not None:
      jax.debug.callback(
          self.progress_fn,
          (iteration, self.inner_iterations, self.max_iterations, state)
      )

    return state

  @property
  def norm_error(self) -> Tuple[int]:  # noqa: D102
    return self._norm_error,

  def init_state(
      self, ot_prob: quadratic_problem.QuadraticProblem,
      init: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]
  ) -> LRGWState:
    """Return the initial state of the loop."""
    q, r, g = init
    return LRGWState(
        q=q,
        r=r,
        g=g,
        gamma=self.gamma,
        costs=-jnp.ones(self.outer_iterations),
        errors=-jnp.ones(self.outer_iterations),
        crossed_threshold=False,
    )

  def output_from_state(
      self, ot_prob: quadratic_problem.QuadraticProblem, state: LRGWState
  ) -> LRGWOutput:
    """Create an output from a loop state.

    Args:
      ot_prob: the transport problem.
      state: GW state.

    Returns:
      A LRGWOutput.
    """
    it = jnp.sum(state.errors != -1.0) * self.inner_iterations
    converged = self._converged(state, it)
    return LRGWOutput(
        q=state.q,
        r=state.r,
        g=state.g,
        ot_prob=ot_prob,
        costs=state.costs,
        errors=state.errors,
        epsilon=self.epsilon,
        inner_iterations=self.inner_iterations,
        converged=converged,
    )

  def _converged(self, state: LRGWState, iteration: int) -> bool:

    def conv_crossed(prev_err: float, curr_err: float) -> bool:
      return jnp.logical_and(
          prev_err < self.threshold, curr_err < self.threshold
      )

    def conv_not_crossed(prev_err: float, curr_err: float) -> bool:
      return jnp.logical_and(curr_err < prev_err, curr_err < self.threshold)

    # for convergence error, we consider 2 possibilities:
    # 1. we either crossed the convergence threshold; in this case we require
    #   that the previous error was also below the threshold
    # 2. we haven't crossed the threshold; in this case, we can be below or
    #   above the threshold:
    #     if we're above, we wait until we reach the convergence threshold and
    #     then, the above condition applies
    #     if we're below and we improved w.r.t. the previous iteration,
    #     we have converged; otherwise we continue, since we may be stuck
    #     in a local minimum (e.g., during the initial iterations)

    it = iteration // self.inner_iterations
    return jax.lax.cond(
        state.crossed_threshold, conv_crossed, conv_not_crossed,
        state.errors[it - 2], state.errors[it - 1]
    )

  def _diverged(self, state: LRGWState, iteration: int) -> bool:
    it = iteration // self.inner_iterations - 1
    is_not_finite = jnp.logical_and(
        jnp.logical_not(jnp.isfinite(state.errors[it])),
        jnp.logical_not(jnp.isfinite(state.costs[it]))
    )
    # `jnp.inf` is used if `it < self.min_iterations`
    return jnp.logical_and(it >= self.min_iterations, is_not_finite)


def run(
    ot_prob: quadratic_problem.QuadraticProblem,
    solver: LRGromovWasserstein,
    init: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
) -> LRGWOutput:
  """Run loop of the solver, outputting a state upgraded to an output."""
  out = sinkhorn.iterations(ot_prob, solver, init)
  out = out.set_cost(
      ot_prob, lse_mode=solver.lse_mode, use_danskin=solver.use_danskin
  )
  return out.set(ot_prob=ot_prob)


def dykstra_solution_error(
    q: jnp.ndarray, r: jnp.ndarray, ot_prob: quadratic_problem.QuadraticProblem,
    norm_error: Tuple[int, ...]
) -> jnp.ndarray:
  """Compute solution error.

  Since only balanced case is available for LR, this is marginal deviation.

  Args:
    q: first factor of solution.
    r: second factor of solution.
    ot_prob: linear problem.
    norm_error: int, p-norm used to compute error.

  Returns:
    one or possibly many numbers quantifying deviation to true marginals.
  """
  norm_error = jnp.array(norm_error)
  # Update the error
  err = jnp.sum(
      jnp.abs(jnp.sum(q, axis=1) - ot_prob.a) ** norm_error[:, None], axis=1
  ) ** (1.0 / norm_error)
  err += jnp.sum(
      jnp.abs(jnp.sum(r, axis=1) - ot_prob.b) ** norm_error[:, None], axis=1
  ) ** (1.0 / norm_error)
  err += jnp.sum(
      jnp.abs(jnp.sum(q, axis=0) - jnp.sum(r, axis=0)) ** norm_error[:, None],
      axis=1
  ) ** (1.0 / norm_error)

  return err


def _linearized_geometry(
    prob: quadratic_problem.QuadraticProblem,
    *,
    q: jnp.ndarray,
    r: jnp.ndarray,
    g: jnp.ndarray,
) -> low_rank.LRCGeometry:
  inv_sqrt_g = 1.0 / jnp.sqrt(g[None, :])

  # TODO(michalk8): below is for squared loss, handle KL loss in the future;
  # will need to be updated in many other places as well
  tmp1 = -4.0 * prob.geom_xx.apply_cost(q, axis=1) * inv_sqrt_g
  tmp2 = prob.geom_yy.apply_cost(r, axis=1) * inv_sqrt_g
  return low_rank.LRCGeometry(tmp1, tmp2)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Callable, Tuple

import jax
import jax.numpy as jnp

import optax

__all__ = ["lbfgs"]

# see https://optax.readthedocs.io/en/stable/_collections/examples/lbfgs.html


def run_opt(
    opt: optax.GradientTransformationExtraArgs,
    x_init: jnp.ndarray,
    fun: Callable[[jnp.ndarray], jnp.ndarray],
    max_iter: int,
    tol: float,
) -> Tuple[jnp.ndarray, optax.OptState]:
  """Runs an optimization algorithm on a function.

  Args:
    opt: An instance of an optax optimizer.
    x_init: Initial point to start optimization.
    fun: The function to minimize.
    max_iter: Maximum number of iterations.
    tol: Tolerance for convergence, measured as the norm of the gradient.

  Returns:
    Final optimization variable obtained after running the optimization.
  """
  value_and_grad_fun = optax.value_and_grad_from_state(fun)

  def step(carry):
    params, state = carry
    value, grad = value_and_grad_fun(params, state=state)
    updates, state = opt.update(
        grad, state, params, value=value, grad=grad, value_fn=fun
    )
    params = optax.apply_updates(params, updates)
    return params, state

  def continuing_criterion(carry):
    _, state = carry
    iter_num = optax.tree.get(state, "count")
    grad = optax.tree.get(state, "grad")
    err = optax.tree.norm(grad)
    return (iter_num == 0) | ((iter_num < max_iter) & (err >= tol))

  init_carry = (x_init, opt.init(x_init))
  final_params, final_state = jax.lax.while_loop(
      continuing_criterion, step, init_carry
  )
  return final_params, final_state


def lbfgs(
    fun: Callable[[jnp.ndarray], jnp.ndarray],
    x_init: jnp.ndarray,
    max_iter: int = 100,
    tol: float = 1e-4,
    **kwargs: Any,
) -> Tuple[jnp.ndarray, optax.OptState]:
  """Runs optax's L-BFGS optimization on function.

  Args:
    fun: The function to minimize.
    x_init: Initial point to start optimization.
    max_iter: Maximum number of iterations.
    tol: Tolerance for convergence.
    kwargs: Keyword arguments for :func:`optax.lbfgs`.

  Returns:
    Final optimization variable obtained after running L-BFGS and state.
  """
  opt = optax.lbfgs(**kwargs)
  return run_opt(opt, x_init, fun, max_iter=max_iter, tol=tol)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
from typing import Any, Callable, Dict, Optional, Tuple

import jax
import jax.numpy as jnp
import jax.tree_util as jtu
import numpy as np

from ott.geometry import costs

try:
  import matplotlib as mpl
  import matplotlib.pyplot as plt
except ImportError:
  mpl = plt = None

__all__ = ["DualPotentials"]

PotentialFn = Callable[[jax.Array], jax.Array]


@jtu.register_static
@dataclasses.dataclass(frozen=True, repr=False)
class DualPotentials:
  r"""The Kantorovich dual potential functions :math:`f` and :math:`g`.

  :math:`f` and :math:`g` are a pair of functions, candidates for the dual
  OT Kantorovich problem, supposedly optimal for a given pair of measures.

  Args:
    f: The first dual potential function.
    g: The second dual potential function.
    cost_fn: The cost function used to solve the OT problem.
  """
  f: Optional[PotentialFn]
  g: Optional[PotentialFn]
  cost_fn: costs.CostFn

  def transport(self, vec: jnp.ndarray, forward: bool = True) -> jnp.ndarray:
    r"""Transport ``vec`` according to Gangbo-McCann Brenier :cite:`brenier:91`.

    Uses Proposition 1.15 from :cite:`santambrogio:15` to compute an OT map when
    applying the inverse gradient of cost.

    When the cost is a general cost, the operator uses the
    :meth:`~ott.geometry.costs.CostFn.twist_operator` associated of the
    corresponding :class:`~ott.geometry.costs.CostFn`.

    When the cost is a translation invariant :class:`~ott.geometry.costs.TICost`
    cost, :math:`c(x,y)=h(x-y)`, and the twist operator translates to the
    application of the convex conjugate of :math:`h` to the
    gradient of the dual potentials, namely
    :math:`x- (\nabla h^*)\circ \nabla f(x)` for the forward map,
    where :math:`h^*` is the Legendre transform of :math:`h`. For instance,
    in the case :math:`h(\cdot) = \|\cdot\|^2, \nabla h(\cdot) = 2 \cdot\,`,
    one has :math:`h^*(\cdot) = \|.\|^2 / 4`, and therefore
    :math:`\nabla h^*(\cdot) = 0.5 \cdot\,`.

    Args:
      vec: Points to transport, array of shape ``[n, d]``.
      forward: Whether to transport the points from source to the target
        distribution or vice-versa.

    Returns:
      The transported points.
    """
    vec = jnp.atleast_2d(vec)
    twist_op = jax.vmap(self.cost_fn.twist_operator, in_axes=[0, 0, None])
    if forward:
      return twist_op(vec, self._grad_f(vec), False)
    return twist_op(vec, self._grad_g(vec), True)

  def distance(self, src: jnp.ndarray, tgt: jnp.ndarray) -> float:
    r"""Evaluate Wasserstein distance between samples using dual potentials.

    This uses direct estimation of potentials against measures when dual
    functions are provided in usual form. This expression is valid for any
    cost function.

    Args:
      src: Samples from the source distribution, array of shape ``[n, d]``.
      tgt: Samples from the target distribution, array of shape ``[m, d]``.

    Returns:
      Wasserstein distance using specified cost function.
    """
    src, tgt = jnp.atleast_2d(src), jnp.atleast_2d(tgt)
    f, g = jax.vmap(self.f), jax.vmap(self.g)
    return jnp.mean(f(src)) + jnp.mean(g(tgt))

  @property
  def _grad_f(self) -> Callable[[jnp.ndarray], jnp.ndarray]:
    """Vectorized gradient of the potential function :attr:`f`."""
    assert self.f is not None, "The `f` potential is not computed."
    return jax.vmap(jax.grad(self.f, argnums=0))

  @property
  def _grad_g(self) -> Callable[[jnp.ndarray], jnp.ndarray]:
    """Vectorized gradient of the potential function :attr:`g`."""
    assert self.g is not None, "The `g` potential is not computed."
    return jax.vmap(jax.grad(self.g, argnums=0))

  def plot_ot_map(
      self,
      source: jnp.ndarray,
      target: jnp.ndarray,
      samples: Optional[jnp.ndarray] = None,
      forward: bool = True,
      ax: Optional["plt.Axes"] = None,
      scatter_kwargs: Optional[Dict[str, Any]] = None,
      legend_kwargs: Optional[Dict[str, Any]] = None,
  ) -> Tuple["plt.Figure", "plt.Axes"]:
    """Plot data and learned optimal transport map.

    Args:
      source: samples from the source measure
      target: samples from the target measure
      samples: extra samples to transport, either ``source`` (if ``forward``) or
        ``target`` (if not ``forward``) by default.
      forward: use the forward map from the potentials if ``True``,
        otherwise use the inverse map.
      ax: axis to add the plot to
      scatter_kwargs: additional kwargs passed into
        :meth:`~matplotlib.axes.Axes.scatter`
      legend_kwargs: additional kwargs passed into
        :meth:`~matplotlib.axes.Axes.legend`

    Returns:
      Figure and axes.
    """
    import matplotlib.pyplot as plt

    if scatter_kwargs is None:
      scatter_kwargs = {"alpha": 0.5}
    if legend_kwargs is None:
      legend_kwargs = {
          "ncol": 3,
          "loc": "upper center",
          "bbox_to_anchor": (0.5, -0.05),
          "edgecolor": "k"
      }

    if ax is None:
      fig = plt.figure(facecolor="white")
      ax = fig.add_subplot(111)
    else:
      fig = ax.get_figure()

    # plot the source and target samples
    if forward:
      label_transport = r"$\nabla f(source)$"
      source_color, target_color = "#1A254B", "#A7BED3"
    else:
      label_transport = r"$\nabla g(target)$"
      source_color, target_color = "#A7BED3", "#1A254B"

    ax.scatter(
        source[:, 0],
        source[:, 1],
        color=source_color,
        label="source",
        **scatter_kwargs,
    )
    ax.scatter(
        target[:, 0],
        target[:, 1],
        color=target_color,
        label="target",
        **scatter_kwargs,
    )

    # plot the transported samples
    samples = (source if forward else target) if samples is None else samples
    transported_samples = self.transport(samples, forward=forward)
    ax.scatter(
        transported_samples[:, 0],
        transported_samples[:, 1],
        color="#F2545B",
        label=label_transport,
        **scatter_kwargs,
    )

    for i in range(samples.shape[0]):
      ax.arrow(
          samples[i, 0],
          samples[i, 1],
          transported_samples[i, 0] - samples[i, 0],
          transported_samples[i, 1] - samples[i, 1],
          color=[0.5, 0.5, 1],
          alpha=0.3,
      )

    ax.legend(**legend_kwargs)
    return fig, ax

  def plot_potential(
      self,
      forward: bool = True,
      quantile: float = 0.05,
      kantorovich: bool = True,
      ax: Optional["mpl.axes.Axes"] = None,
      x_bounds: Tuple[float, float] = (-6, 6),
      y_bounds: Tuple[float, float] = (-6, 6),
      num_grid: int = 50,
      contourf_kwargs: Optional[Dict[str, Any]] = None,
  ) -> Tuple["mpl.figure.Figure", "mpl.axes.Axes"]:
    r"""Plot the potential.

    Args:
      forward: use the forward map from the potentials
        if ``True``, otherwise use the inverse map
      quantile: quantile to filter the potentials with
      kantorovich: whether to plot the Kantorovich potential
      ax: axis to add the plot to
      x_bounds: x-axis bounds of the plot
        :math:`(x_{\text{min}}, x_{\text{max}})`
      y_bounds: y-axis bounds of the plot
        :math:`(y_{\text{min}}, y_{\text{max}})`
      num_grid: number of points to discretize the domain into a grid
        along each dimension
      contourf_kwargs: additional kwargs passed into
        :meth:`~matplotlib.axes.Axes.contourf`

    Returns:
      Figure and axes.
    """
    import matplotlib.pyplot as plt

    if contourf_kwargs is None:
      contourf_kwargs = {}

    ax_specified = ax is not None
    if not ax_specified:
      fig, ax = plt.subplots(figsize=(6, 6), facecolor="white")
    else:
      fig = ax.get_figure()

    x1 = jnp.linspace(*x_bounds, num=num_grid)
    x2 = jnp.linspace(*y_bounds, num=num_grid)
    X1, X2 = jnp.meshgrid(x1, x2)
    X12flat = jnp.hstack((X1.reshape(-1, 1), X2.reshape(-1, 1)))
    Zflat = jax.vmap(self.f if forward else self.g)(X12flat)
    if kantorovich:
      Zflat = 0.5 * (jnp.linalg.norm(X12flat, axis=-1) ** 2) - Zflat
    Zflat = np.asarray(Zflat)
    vmin, vmax = np.quantile(Zflat, [quantile, 1.0 - quantile])
    Zflat = Zflat.clip(vmin, vmax)
    Z = Zflat.reshape(X1.shape)

    CS = ax.contourf(X1, X2, Z, cmap="Blues", **contourf_kwargs)
    ax.set_xlim(*x_bounds)
    ax.set_ylim(*y_bounds)
    fig.colorbar(CS, ax=ax)
    if not ax_specified:
      fig.tight_layout()
    ax.set_title(r"$f$" if forward else r"$g$")
    return fig, ax
